__all__ = ('PollQuestion',)

from scarletio import RichAttributeErrorBaseType

from ...utils import sanitize_mentions

from .fields import parse_text, put_text, validate_text


class PollQuestion(RichAttributeErrorBaseType):
    """
    Represents a poll's question.
    
    Attributes
    ----------
    text : `None`, `str`
        The poll question's text.
    """
    __slots__ = ('text',)
    
    def __new__(cls, *, text = ...):
        """
        Creates a new poll question with the given parameters.
        
        Parameters
        ----------
        text : `None`, `str`, Optional (Keyword only)
            The question's text.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # text
        if text is ...:
            text = None
        else:
            text = validate_text(text)
        
        # Construct
        self = object.__new__(cls)
        self.text = text
        return self
    
    
    def __len__(self):
        """Returns the poll question's length."""
        length = 0
        
        # text
        text = self.text
        if (text is not None):
            length += len(text)
        
        return length
    
    
    def __bool__(self):
        """Returns whether the poll question has any fields set."""
        # text
        if self.text is not None:
            return True
        
        return False
    
    
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        # text
        text = self.text
        if (text is not None):
            repr_parts.append(' text = ')
            repr_parts.append(repr(text))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the poll question's hash value."""
        hash_value = 0
        
        # text
        text = self.text
        if (text is not None):
            hash_value ^= hash(text)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two poll questions are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # text
        if self.text != other.text:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new poll question instance from the given json data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Data to create poll question from.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.text = parse_text(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Returns the poll question as a json serializable representation.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_text(self.text, data, defaults)
        return data
    
    
    def clean_copy(self, guild = None):
        """
        Creates a clean copy of the poll question by removing the mentions in it's contents.
        
        Parameters
        ----------
        guild : ``None | Guild`` = `None`, Optional
            The respective guild as a context to look up guild specific names of entities.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.text = sanitize_mentions(self.text, guild)
        return new
    
    
    def copy(self):
        """
        Copies the poll question.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.text = self.text
        return new
    
    
    def copy_with(self, *, text = ...):
        """
        Copies the poll question with the given parameters.
        
        Parameters
        ----------
        text : `None`, `str`, Optional (Keyword only)
            The question's text.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # text
        if text is ...:
            text = self.text
        else:
            text = validate_text(text)
        
        # Construct
        new = object.__new__(type(self))
        new.text = text
        return new
    
    
    @property
    def contents(self):
        """
        Returns the contents of the poll question.
        
        Returns
        -------
        contents : `list<str>`
        """
        return [*self.iter_contents()]
    
    
    def iter_contents(self):
        """
        Iterates over the contents of the poll question.
        
        This method is an iterable generator.
        
        Yields
        ------
        content : `str`
        """
        text = self.text
        if (text is not None):
            yield text



__all__ = ()

from ...application_command import ApplicationCommandPermissionOverwrite
from ...emoji import create_partial_emoji_from_id, create_unicode_emoji

from ..audit_log_change import AuditLogChange
from ..audit_log_change.flags import FLAG_HAS_AFTER, FLAG_HAS_BEFORE


def change_deserializer_modification(conversion, change_data):
    """
    Default deserializer used for regular modification entries.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    conversion : ``AuditLogEntryChangeConversion``
        The conversion to pull its field of.
    change_data : `dict<str, object>`
        Audit log change to serialize.
    
    Yields
    ------
    change : ``AuditLogChange``
    """
    value_deserializer = conversion.value_deserializer
    flags = 0
    
    try:
        value = change_data['old_value']
    except KeyError:
        before = None
    else:
        if value_deserializer is None:
            before = value
        else:
            before = value_deserializer(value)
        
        flags |= FLAG_HAS_BEFORE
    
    try:
        value = change_data['new_value']
    except KeyError:
        after = None
    else:
        if value_deserializer is None:
            after = value
        else:
            after = value_deserializer(value)
        
        flags |= FLAG_HAS_AFTER
    
    if not flags:
        return
    
    yield AuditLogChange.from_fields(conversion.field_name, flags, before, after)


def change_deserializer_addition_and_removal(conversion, change_data):
    """
    Serializer for additions and removals.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    conversion : ``AuditLogEntryChangeConversion``
        The conversion to pull its field of.
    change_data : `dict<str, object>`
        Audit log change to serialize.
    
    Yields
    ------
    change : ``AuditLogChange``
    """
    value_deserializer = conversion.value_deserializer
    flags = 0
    
    key = change_data.get('key', None)
    
    if key == conversion.get_field_key(0):
        try:
            value = change_data['new_value']
        except KeyError:
            before = None
        else:
            if value_deserializer is None:
                before = value
            else:
                before = value_deserializer(value)
            
            flags |= FLAG_HAS_BEFORE
        
        after = None
    
    else:
        before = None
        
        try:
            value = change_data['new_value']
        except KeyError:
            after = None
        else:
            if value_deserializer is None:
                after = value
            else:
                after = value_deserializer(value)
            
            flags |= FLAG_HAS_AFTER
    
    
    if not flags:
        return
    
    yield AuditLogChange.from_fields(conversion.field_name, flags, before, after)


def change_deserializer_deprecation(conversion, change_data):
    """
    Serializer for deprecated fields. Does nothing.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    conversion : ``AuditLogEntryChangeConversion``
        The conversion to pull its field of.
    change_data : `dict<str, object>`
        Audit log change to serialize.
    
    Yields
    ------
    change : ``AuditLogChange``
    """
    return
    yield


def _value_deserializer_emoji_custom(value):
    """
    Default get converter returned for custom emojis.
    
    Parameters
    ----------
    value : `None | str`
        Emoji identifier string.
    
    Returns
    -------
    emoji : ``None | Emoji``
    """
    if value is None:
        emoji = None
    else:
        emoji = create_partial_emoji_from_id(int(value))
    
    return emoji


def _value_deserializer_emoji_unicode(value):
    """
    Default get converter returned for unicode emojis.
    
    Parameters
    ----------
    value : `None | str`
        Emoji unicode string.
    
    Returns
    -------
    emoji : ``None | Emoji``
    """
    if value is None:
        emoji = None
    else:
        emoji = create_unicode_emoji(value)
    
    return emoji


def change_deserializer_flattened_emoji(conversion, change_data):
    """
    Serializer for additions and removals.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    conversion : ``AuditLogEntryChangeConversion``
        The conversion to pull its field of.
    change_data : `dict<str, object>`
        Audit log change to serialize.
    
    Yields
    ------
    change : ``AuditLogChange``
    """
    key = change_data.get('key', None)
    value_deserializer = _value_deserializer_emoji_custom if key == 'emoji_id' else _value_deserializer_emoji_unicode
    
    flags = 0
    
    try:
        value = change_data['old_value']
    except KeyError:
        before = None
    else:
        before = value_deserializer(value)
        flags |= FLAG_HAS_BEFORE
    
    try:
        value = change_data['new_value']
    except KeyError:
        after = None
    else:
        after = value_deserializer(value)
        flags |= FLAG_HAS_AFTER
    
    if not flags:
        return
    
    yield AuditLogChange.from_fields(conversion.field_name, flags, before, after)


def change_deserializer_application_command_permission_overwrite(conversion, change_data):
    """
    Serializer an application command permission overwrite.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    conversion : ``AuditLogEntryChangeConversion``
        The conversion to pull its field of.
    change_data : `dict<str, object>`
        Audit log change to serialize.
    
    Yields
    ------
    change : ``AuditLogChange``
    """
    flags = 0
    
    try:
        value = change_data['old_value']
    except KeyError:
        before = None
    else:
        if value is None:
            before = None
        else:
            before = (ApplicationCommandPermissionOverwrite.from_data(value),)
        
        flags |= FLAG_HAS_BEFORE
    
    try:
        value = change_data['new_value']
    except KeyError:
        after = None
    else:
        if value is None:
            after = None
        else:
            after = (ApplicationCommandPermissionOverwrite.from_data(value),)
        flags |= FLAG_HAS_AFTER
    
    if not flags:
        return
    
    yield AuditLogChange.from_fields(conversion.field_name, flags, before, after)



__all__ = ()

from ...onboarding import OnboardingMode, OnboardingPrompt
from ...onboarding.onboarding_screen.fields import (
    validate_default_channel_ids, validate_enabled, validate_mode, validate_prompts
)

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..audit_log_entry_change_conversion.change_deserializers import change_deserializer_deprecation
from ..conversion_helpers.converters import value_deserializer_ids, value_serializer_ids


# ---- below_requirements ---- 

BELOW_REQUIREMENTS_CONVERSION_IGNORED = AuditLogEntryChangeConversion(
    ('below_requirements',),
    '',
    change_deserializer = change_deserializer_deprecation,
)


# ---- default_channel_ids ----

DEFAULT_CHANNEL_IDS_CONVERSION = AuditLogEntryChangeConversion(
    ('default_channel_ids',),
    'default_channel_ids',
    value_deserializer = value_deserializer_ids,
    value_serializer = value_serializer_ids,
    value_validator = validate_default_channel_ids,
)


# ---- enabled ----

ENABLED_CONVERSION = AuditLogEntryChangeConversion(
    ('enabled',),
    'enabled',
    value_validator = validate_enabled,
)


@ENABLED_CONVERSION.set_value_deserializer
def enabled_value_deserializer(value):
    if value is None:
        value = True
    return value


# ---- mode ----

MODE_CONVERSION = AuditLogEntryChangeConversion(
    ('mode',),
    'mode',
    value_validator = validate_mode,
)


@MODE_CONVERSION.set_value_deserializer
def mode_value_deserializer(value):
    return OnboardingMode(value)


@MODE_CONVERSION.set_value_serializer
def mode_value_serializer(value):
    return value.value


# ---- prompts ----

PROMPTS_CONVERSION = AuditLogEntryChangeConversion(
    ('prompts',),
    'prompts',
    value_validator = validate_prompts,
)


@PROMPTS_CONVERSION.set_value_deserializer
def prompts_value_deserializer(value):
    if value is None:
        pass
    elif (not value):
        value = None
    else:
        value = (*(OnboardingPrompt.from_data(data) for data in value),)
    
    return value


@PROMPTS_CONVERSION.set_value_serializer
def prompts_value_serializer(value):
    if value is None:
        value = []
    else:
        value = [prompt.to_data(defaults = True, include_internals = True) for prompt in value]
    
    return value


# ---- Construct ----

ONBOARDING_SCREEN_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    BELOW_REQUIREMENTS_CONVERSION_IGNORED,
    DEFAULT_CHANNEL_IDS_CONVERSION,
    ENABLED_CONVERSION,
    MODE_CONVERSION,
    PROMPTS_CONVERSION,
)



__all__ = ()

from functools import partial as partial_func

from ...application import Application, ApplicationIntegrationType
from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, flag_parser_factory, force_string_parser_factory,
    nullable_functional_parser_factory, nullable_object_array_parser_factory, nullable_string_parser_factory,
    preinstanced_array_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_optional_putter_factory, entity_id_putter_factory,
    force_string_putter_factory, nullable_entity_array_optional_putter_factory,
    nullable_functional_optional_putter_factory, nullable_string_putter_factory, preinstanced_optional_putter_factory,
    preinstanced_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, flag_validator_factory,
    nullable_object_array_validator_factory, nullable_string_validator_factory, preinstanced_array_validator_factory,
    preinstanced_validator_factory
)
from ...guild import Guild
from ...localization.helpers import localized_dictionary_builder
from ...localization.utils import build_locale_dictionary, destroy_locale_dictionary
from ...permission import Permission
from ...utils import is_valid_application_command_name

from ..application_command_option import ApplicationCommandOption

from .constants import DESCRIPTION_LENGTH_MAX, DESCRIPTION_LENGTH_MIN, NAME_LENGTH_MAX, NAME_LENGTH_MIN, OPTIONS_MAX
from .preinstanced import (
    ApplicationCommandHandlerType, ApplicationCommandIntegrationContextType, ApplicationCommandTargetType
)


# application_id

parse_application_id = entity_id_parser_factory('application_id')
put_application_id = entity_id_putter_factory('application_id')
validate_application_id = entity_id_validator_factory('application_id', Application)

# description

parse_description = nullable_string_parser_factory('description')
put_description = nullable_string_putter_factory('description')
validate_description = nullable_string_validator_factory(
    'description', DESCRIPTION_LENGTH_MIN, DESCRIPTION_LENGTH_MAX
)

# description_localizations

parse_description_localizations = nullable_functional_parser_factory(
    'description_localizations', build_locale_dictionary
)
put_description_localizations = nullable_functional_optional_putter_factory(
    'description_localizations', destroy_locale_dictionary
)
validate_description_localizations = partial_func(
    localized_dictionary_builder, parameter_name = 'description_localizations'
)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id = entity_id_optional_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', Guild)


# handler_type

parse_handler_type = preinstanced_parser_factory(
    'handler', ApplicationCommandHandlerType, ApplicationCommandHandlerType.none
)
put_handler_type = preinstanced_optional_putter_factory('handler', ApplicationCommandHandlerType.none)
validate_handler_type = preinstanced_validator_factory('handler_type', ApplicationCommandHandlerType)


# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('id')

# integration_context_types

parse_integration_context_types = preinstanced_array_parser_factory(
    'contexts', ApplicationCommandIntegrationContextType
)


def put_integration_context_types(integration_context_types, data, defaults):
    """
    Puts the `integration_context_types`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    integration_context_types : `None | tuple<ApplicationCommandIntegrationContextType>`
        The places where the application command shows up.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if integration_context_types is None:
        raw = None
    else:
        raw = [preinstanced.value for preinstanced in integration_context_types]
    
    data['contexts'] = raw
    return data


validate_integration_context_types = preinstanced_array_validator_factory(
    'integration_context_types', ApplicationCommandIntegrationContextType
)


# integration_types

parse_integration_types = preinstanced_array_parser_factory('integration_types', ApplicationIntegrationType)


def put_integration_types(integration_types, data, defaults):
    """
    Puts the `integration_types`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    integration_types : `None | tuple<ApplicationIntegrationType>`
        The options where the application command can be integrated to.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if integration_types is None:
        raw = None
    else:
        raw = [preinstanced.value for preinstanced in integration_types]
    
    data['integration_types'] = raw
    return data


validate_integration_types = preinstanced_array_validator_factory('integration_types', ApplicationIntegrationType)


# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')


def validate_name(name):
    """
    Validates the given application command's name.
    
    Parameters
    ----------
    name : `None`, `str`
        The application command name to validate.
    
    Returns
    -------
    name : `str`
        The validated name.
    
    Raises
    ------
    TypeError
        - If `name`'s type is incorrect.
    ValueError
        - If `name`'s value is incorrect.
    """
    if name is None:
        return ''
    
    if not isinstance(name, str):
        raise TypeError(
            f'`name` can be `None`, `str`, got {type(name).__name__} ;{name!r}.'
        )
    
    name_length = len(name)
    if (
        name_length and (
            (name_length < NAME_LENGTH_MIN) or
            (name_length > NAME_LENGTH_MAX)
        )
    ):
        raise ValueError(
            f'`name` length can be >= {NAME_LENGTH_MIN} and '
            f'<= {NAME_LENGTH_MAX}, got {name_length}; name = {name!r}'
        )
    
    if not is_valid_application_command_name(name):
        raise ValueError(
            f'`name` contains unexpected character(s) (that are not valid as an application command name); '
            f'got {name!r}.'
        )
    
    return name


# name_localizations

parse_name_localizations = nullable_functional_parser_factory(
    'name_localizations', build_locale_dictionary
)
put_name_localizations = nullable_functional_optional_putter_factory(
    'name_localizations', destroy_locale_dictionary
)
validate_name_localizations = partial_func(
    localized_dictionary_builder, parameter_name = 'name_localizations'
)

# nsfw

parse_nsfw = bool_parser_factory('nsfw', False)
put_nsfw = bool_optional_putter_factory('nsfw', False)
validate_nsfw = bool_validator_factory('nsfw', False)

# options

parse_options = nullable_object_array_parser_factory('options', ApplicationCommandOption)
put_options = nullable_entity_array_optional_putter_factory('options', ApplicationCommandOption)
_pre_validate_options = nullable_object_array_validator_factory('options', ApplicationCommandOption)


def validate_options(options):
    """
    Validates the given application command option `options` field.
    
    Parameters
    ----------
    options : `None`, `iterable` of ``ApplicationCommandOption``
        The parameters or sub-commands of the command option. It's length can be in range [0:25].
    
    Returns
    -------
    options : `None`, `tuple` of ``ApplicationCommandOption``
    
    Returns
    -------
    TypeError
        - If `options`'s type is incorrect.
    ValueError
        - If `options`'s value is incorrect.
        - If `options` is not applicable for the given type.
    """
    options = _pre_validate_options(options)
    if (options is not None):
        options = options[:OPTIONS_MAX]
    
    return options

# required_permissions

parse_required_permissions = flag_parser_factory('default_member_permissions', Permission)


def put_required_permissions(required_permissions, data, defaults):
    """
    Puts the given application command's required permissions into the given data.
    
    Parameters
    ----------
    required_permissions : ``Permission``
        The required permissions a user should have to use the command.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if required_permissions:
        required_permissions = format(required_permissions, 'd')
    else:
        required_permissions = None
    data['default_member_permissions'] = required_permissions
    return data


validate_required_permissions = flag_validator_factory('required_permissions', Permission)

# target_type

parse_target_type = preinstanced_parser_factory(
    'type', ApplicationCommandTargetType, ApplicationCommandTargetType.none
)
put_target_type = preinstanced_putter_factory('type')
validate_target_type = preinstanced_validator_factory('target_type', ApplicationCommandTargetType)

# version

parse_version = entity_id_parser_factory('version')
put_version = entity_id_optional_putter_factory('version')
validate_version = entity_id_validator_factory('version')



__all__ = ()

from datetime import datetime as DateTime

from ...channel import Channel
from ...field_parsers import (
    default_entity_parser_factory, entity_id_array_parser_factory, entity_id_parser_factory,
    force_string_parser_factory, int_parser_factory, nullable_date_time_parser_factory, nullable_string_parser_factory,
    preinstanced_parser_factory, nullable_entity_parser_factory
)
from ...field_putters import (
    default_entity_putter_factory, entity_id_optional_putter_factory, entity_id_putter_factory,
    force_string_putter_factory, int_putter_factory,  nullable_date_time_optional_putter_factory,
    optional_entity_id_array_optional_putter_factory, nullable_string_optional_putter_factory,
    preinstanced_optional_putter_factory, preinstanced_putter_factory
)
from ...field_validators import (
    default_entity_validator_factory, entity_id_array_validator_factory, entity_id_validator_factory,
    force_string_validator_factory, int_conditional_validator_factory, nullable_date_time_validator_factory,
    nullable_string_validator_factory, preinstanced_validator_factory, nullable_entity_validator_factory,
    nullable_object_array_validator_factory
)
from ...user import ClientUserBase, User, ZEROUSER
from ...utils import datetime_to_id, id_to_datetime

from ..schedule import Schedule
from ..scheduled_event_occasion_overwrite import ScheduledEventOccasionOverwrite
from ..scheduled_event_entity_metadata import ScheduledEventEntityMetadataBase, ScheduledEventEntityMetadataLocation

from .constants import DESCRIPTION_LENGTH_MAX, DESCRIPTION_LENGTH_MIN, NAME_LENGTH_MAX, NAME_LENGTH_MIN
from .preinstanced import PrivacyLevel, ScheduledEventEntityType, ScheduledEventStatus


# occasion_overwrites


def parse_occasion_overwrites(data):
    """
    Parses occasion_overwrites out from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    occasion_overwrites : `None | tuple<DateTime>`
    """
    occasion_overwrite_datas = data.get('guild_scheduled_event_exceptions', None)
    if (occasion_overwrite_datas is not None) and occasion_overwrite_datas:
        return tuple(sorted(
            ScheduledEventOccasionOverwrite.from_data(occasion_overwrite_data) for occasion_overwrite_data in occasion_overwrite_datas
        ))


def put_occasion_overwrites(occasion_overwrites, data, defaults, *, scheduled_event_id = 0):
    """
    Serialises the occasion_overwrites into the given data.
    
    Parameters
    ----------
    occasion_overwrites : `None | tuple<DateTime>`
        The cancelled occurrences.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    scheduled_event_id : `int` = `0`, Optional (Keyword only)
        The owner scheduled event's identifier.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (occasion_overwrites is not None):
        occasion_overwrite_datas = []
        
        if (occasion_overwrites is not None):
            if scheduled_event_id:
                scheduled_event_id_serialized = str(scheduled_event_id)
            else:
                scheduled_event_id_serialized = None
            
            for occasion_overwrite in occasion_overwrites:
                occasion_overwrite_data = occasion_overwrite.to_data(defaults = defaults, include_internals = True)
                occasion_overwrite_data['event_id'] = scheduled_event_id_serialized
                
                occasion_overwrite_datas.append(occasion_overwrite_data)
        
        data['guild_scheduled_event_exceptions'] = occasion_overwrite_datas
    
    return data


validate_occasion_overwrites = nullable_object_array_validator_factory(
    'occasion_overwrites', ScheduledEventOccasionOverwrite, ordered = True
)


# channel_id

parse_channel_id = entity_id_parser_factory('channel_id')
put_channel_id = entity_id_optional_putter_factory('channel_id')
validate_channel_id = entity_id_validator_factory('channel_id', Channel)


# creator

parse_creator = default_entity_parser_factory('creator', User, default = ZEROUSER)
put_creator = default_entity_putter_factory('creator', ClientUserBase, ZEROUSER, force_include_internals = True)
validate_creator = default_entity_validator_factory('creator', ClientUserBase, default = ZEROUSER)


# description

parse_description = nullable_string_parser_factory('description')
put_description = nullable_string_optional_putter_factory('description')
validate_description = nullable_string_validator_factory('description', DESCRIPTION_LENGTH_MIN, DESCRIPTION_LENGTH_MAX)


# end

parse_end = nullable_date_time_parser_factory('scheduled_end_time')
put_end = nullable_date_time_optional_putter_factory('scheduled_end_time')
validate_end = nullable_date_time_validator_factory('end')


# entity_id

parse_entity_id = entity_id_parser_factory('entity_id')
put_entity_id = entity_id_optional_putter_factory('entity_id')
validate_entity_id = entity_id_validator_factory('entity_id')


# entity_metadata

def parse_entity_metadata(data, entity_type):
    """
    Parsers out a scheduled event entity's metadata form the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Scheduled event data.
    
    entity_type : ``ScheduledEventEntityType``
        The scheduled event entity's type.
    
    Returns
    -------
    entity_metadata : ``ScheduledEventEntityMetadataBase``
    """
    entity_metadata_type = entity_type.metadata_type
    entity_metadata_data = data.get('entity_metadata', None)
    
    if entity_metadata_data is None:
        entity_metadata = entity_metadata_type._create_empty()
    else:
        entity_metadata = entity_metadata_type.from_data(entity_metadata_data)
    
    return entity_metadata


def put_entity_metadata(entity_metadata, data, defaults):
    """
    Puts the given scheduled event entity's metadata into the given `data` json serializable object.
    
    Parameters
    ----------
    entity_metadata : ``ScheduledEventEntityMetadataBase``
        Scheduled event entity metadata.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (type(entity_metadata) is not ScheduledEventEntityMetadataBase):
        data['entity_metadata'] = entity_metadata.to_data(defaults = defaults)
    
    return data


# entity_type

parse_entity_type = preinstanced_parser_factory('entity_type', ScheduledEventEntityType, ScheduledEventEntityType.none)
put_entity_type = preinstanced_putter_factory('entity_type')
validate_entity_type = preinstanced_validator_factory('entity_type', ScheduledEventEntityType)


# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id = entity_id_optional_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')


# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('scheduled_event_id')


# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)


# privacy_level

parse_privacy_level = preinstanced_parser_factory('privacy_level', PrivacyLevel, PrivacyLevel.guild_only)
put_privacy_level = preinstanced_putter_factory('privacy_level')
validate_privacy_level = preinstanced_validator_factory('privacy_level', PrivacyLevel)


# schedule

parse_schedule = nullable_entity_parser_factory('recurrence_rule', Schedule)


def put_schedule(schedule, data, defaults, *, start = None):
    """
    Serialises the given schedule.
    
    Parameters
    ----------
    schedule : ``None | Schedule``
        The schedule to serialize.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    start : `None | DateTime` = `None`, Optional (Keyword only)
        Start date to use if schedule does not define it.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (schedule is not None):
        if schedule is None:
            schedule_data = None
        else:
            schedule_data = schedule.to_data(defaults = defaults, start = start)
        
        data['recurrence_rule'] = schedule_data
    
    return data    


validate_schedule = nullable_entity_validator_factory('schedule', Schedule)


# start

parse_start = nullable_date_time_parser_factory('scheduled_start_time')
put_start = nullable_date_time_optional_putter_factory('scheduled_start_time')
validate_start = nullable_date_time_validator_factory('start')


# sku_ids

parse_sku_ids = entity_id_array_parser_factory('sku_ids')
put_sku_ids = optional_entity_id_array_optional_putter_factory('sku_ids')
validate_sku_ids = entity_id_array_validator_factory('sku_ids', NotImplemented, include = 'SKU')


# status

parse_status = preinstanced_parser_factory('status', ScheduledEventStatus, ScheduledEventStatus.none)
put_status = preinstanced_optional_putter_factory('status', ScheduledEventStatus.none)
validate_status = preinstanced_validator_factory('status', ScheduledEventStatus)


# user_count

parse_user_count = int_parser_factory('user_count', 0)
put_user_count = int_putter_factory('user_count')
validate_user_count = int_conditional_validator_factory(
    'user_count',
    0,
    (lambda user_count : user_count >= 0),
    '>= 0',
)


# Target parsers

ENTITY_METADATA_DEFAULT = ScheduledEventEntityMetadataBase._create_empty()


# location

def validate_target_location(location):
    """
    Validates scheduled event target (location).
    
    Parameters
    ----------
    location : `None`, `str`
        The place where the event will take place.
    
    Returns
    -------
    entity_type : ``ScheduledEventEntityType``
        Scheduled event entity type.
    entity_metadata : ``ScheduledEventEntityMetadataBase``
        Scheduled event entity metadata.
    channel_id : `int`
        Scheduled event target channel identifier.
    
    Raises
    ------
    TypeError
        - If `location`'s type is incorrect.
    ValueError
        - If `location`'s value is incorrect.
    """
    entity_metadata = ScheduledEventEntityMetadataLocation(location = location)
    return ScheduledEventEntityType.location, entity_metadata, 0


# voice

def validate_target_voice(voice):
    """
    Validates scheduled event target (voice channel).
    
    Parameters
    ----------
    voice : ``Channel``
        The voice channel where the event will take place.
    
    Returns
    -------
    entity_type : ``ScheduledEventEntityType``
        Scheduled event entity type.
    entity_metadata : ``ScheduledEventEntityMetadataBase``
        Scheduled event entity metadata.
    channel_id : `int`
        Scheduled event target channel identifier.
    
    Raises
    ------
    TypeError
        - If `voice`'s type is incorrect.
    ValueError
        - If `voice`'s value is incorrect.
    """
    channel_id = validate_channel_id(voice)
    return ScheduledEventEntityType.voice, ENTITY_METADATA_DEFAULT, channel_id


# stage

def validate_target_stage(stage):
    """
    Validates scheduled event target (stage channel).
    
    Parameters
    ----------
    stage : ``Channel``
        The stage channel where the event will take place.
    
    Returns
    -------
    entity_type : ``ScheduledEventEntityType``
        Scheduled event entity type.
    entity_metadata : ``ScheduledEventEntityMetadataBase``
        Scheduled event entity metadata.
    channel_id : `int`
        Scheduled event target channel identifier.
    
    Raises
    ------
    TypeError
        - If `stage`'s type is incorrect.
    ValueError
        - If `stage`'s value is incorrect.
    """
    channel_id = validate_channel_id(stage)
    return ScheduledEventEntityType.stage, ENTITY_METADATA_DEFAULT, channel_id


# target

def put_target(target, data, defaults):
    """
    Puts the given scheduled event target into the given `data` json serializable object.
    
    Used when creating  or editing a scheduled events.
    
    Parameters
    ----------
    target : `tuple` (``ScheduledEventEntityType``, ``ScheduledEventEntityMetadataBase``, `int`)
        Scheduled event target.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    entity_type, entity_metadata, channel_id = target
    put_entity_type(entity_type, data, defaults)
    put_entity_metadata(entity_metadata, data, defaults)
    put_channel_id(channel_id, data, defaults)
    return data



__all__ = (
    'create_partial_channel_data', 'create_partial_channel_from_data', 'create_partial_channel_from_id', 
)

from functools import partial as partial_func
from warnings import warn

from scarletio import export, include

from ...core import CHANNELS

from ..channel_metadata.private_group import CHANNEL_METADATA_ICON
from ..channel_metadata.fields import (
    put_applied_tag_ids, put_auto_archive_after, put_bitrate, put_default_forum_layout,
    put_default_sort_order, put_default_thread_auto_archive_after, put_default_thread_reaction_emoji,
    put_default_thread_slowmode, put_flags, put_invitable, put_name, put_nsfw, put_open,
    put_parent_id, put_permission_overwrites, put_position, put_region, put_slowmode,
    put_topic, put_user_limit, put_video_quality_mode, validate_applied_tag_ids,
    validate_auto_archive_after, validate_bitrate, validate_default_forum_layout, validate_default_sort_order,
    validate_default_thread_auto_archive_after, validate_default_thread_reaction_emoji, validate_default_thread_slowmode,
    validate_flags, validate_invitable, validate_name, validate_nsfw, validate_open, validate_parent_id,
    validate_permission_overwrites, validate_position, validate_region, validate_slowmode, validate_topic,
    validate_user_limit, validate_video_quality_mode
)

from .fields import parse_id, put_guild_id, put_id, put_type, validate_type
from .preinstanced import ChannelType


Channel = include('Channel')


def _deprecated__validate_default_thread_reaction(emoji):
    warn(
        '`default_thread_reaction` parameter is deprecated. Please use `default_thread_reaction_emoji` instead.',
        FutureWarning,
        stacklevel = 4,
    )
    return validate_default_thread_reaction_emoji(emoji)


CHANNEL_GUILD_MAIN_FIELD_CONVERTERS = {
    'bitrate': (validate_bitrate, put_bitrate),
    'channel_type': (validate_type, put_type),
    'default_forum_layout': (validate_default_forum_layout, put_default_forum_layout),
    'default_sort_order': (validate_default_sort_order, put_default_sort_order),
    'default_thread_auto_archive_after': (
        validate_default_thread_auto_archive_after, put_default_thread_auto_archive_after
    ),
    'default_thread_reaction': (_deprecated__validate_default_thread_reaction, put_default_thread_reaction_emoji),
    'default_thread_reaction_emoji': (validate_default_thread_reaction_emoji, put_default_thread_reaction_emoji),
    'default_thread_slowmode': (validate_default_thread_slowmode, put_default_thread_slowmode),
    'flags': (validate_flags, put_flags),
    'name': (validate_name, put_name),
    'nsfw': (validate_nsfw, put_nsfw),
    'parent_id': (validate_parent_id, put_parent_id),
    'permission_overwrites': (validate_permission_overwrites, put_permission_overwrites),
    'position': (validate_position, put_position),
    'region': (validate_region, put_region),
    'slowmode': (validate_slowmode, put_slowmode),
    'topic': (validate_topic, put_topic),
    'user_limit': (validate_user_limit, put_user_limit),
    'video_quality_mode': (validate_video_quality_mode, put_video_quality_mode),
}


CHANNEL_PRIVATE_GROUP_FIELD_CONVERTERS = {
    'icon': (
        CHANNEL_METADATA_ICON.validate_icon,
        partial_func(CHANNEL_METADATA_ICON.put_into, as_data = True),
    ),
    'name': (validate_name, put_name),
}


CHANNEL_GUILD_THREAD_FIELD_CONVERTERS = {
    'applied_tags': (validate_applied_tag_ids, put_applied_tag_ids),
    'applied_tag_ids': (validate_applied_tag_ids, put_applied_tag_ids),
    'auto_archive_after': (
        validate_auto_archive_after, partial_func(put_auto_archive_after, flatten_thread_metadata = True)
    ),
    'flags': (validate_flags, put_flags),
    'invitable': (validate_invitable, partial_func(put_invitable, flatten_thread_metadata = True)),
    'name': (validate_name, put_name),
    'open_': (validate_open, partial_func(put_open, flatten_thread_metadata = True)),
    'slowmode': (validate_slowmode, put_slowmode),
}


CHANNEL_GUILD_FIELD_CONVERTERS = {
    **CHANNEL_GUILD_THREAD_FIELD_CONVERTERS,
    **CHANNEL_GUILD_MAIN_FIELD_CONVERTERS,
}


@export
def create_partial_channel_from_data(data, guild_id = 0):
    """
    Creates a partial channel from partial channel data.
    
    Parameters
    ----------
    data : `None`, `dict<str, object>`
        Partial channel data received from Discord.
    guild_id : `int` = `0`, Optional (Keyword only)
        The channel's guild's identifier.
    
    Returns
    -------
    channel : ``None | Channel``
        The created partial channel, or `None`, if no data was received.
    """
    channel_id = parse_id(data)
    
    try:
        return CHANNELS[channel_id]
    except KeyError:
        pass
    
    channel = Channel._from_partial_data(data, channel_id, guild_id)
    CHANNELS[channel_id] = channel
    
    return channel


@export
def create_partial_channel_from_id(channel_id, channel_type, guild_id):
    """
    Creates a new partial channel from the given identifier.
    
    Parameters
    ----------
    channel_id : `int`
        The channel's identifier.
    channel_type : ``ChannelType``
        The channel's type.
    guild_id : `int`
        A guild's identifier of the created channel.
    
    Returns
    -------
    channel : ``Channel``
    """
    try:
        return CHANNELS[channel_id]
    except KeyError:
        pass
    
    channel = Channel._create_empty(channel_id, channel_type, guild_id)
    CHANNELS[channel_id] = channel
    
    return channel


@export
def create_partial_channel_data(channel):
    """
    Creates partial channel data for the given channel.
    
    Parameters
    ----------
    channel : ``Channel``
        The channel to create the partial data from.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    data = {}
    put_id(channel.id, data, True)
    put_guild_id(channel.guild_id, data, True)
    put_type(channel.type, data, True)
    put_name(channel.name, data, True)
    return data


HAS_SLOWMODE = (
    *(
        channel_type for channel_type in ChannelType.INSTANCES.values()
        if channel_type.flags.guild and (channel_type.flags.textual or channel_type.flags.forum)
    ),
)



__all__ = ('abort', 'InteractionAbortedError', 'InteractionResponse',)

from itertools import islice

from scarletio import is_coroutine_generator, skip_ready_cycle

from ...discord.application_command import ApplicationCommandTargetType
from ...discord.client import Client
from ...discord.component import InteractionForm
from ...discord.interaction import InteractionType
from ...discord.message import (
    MessageBuilderBase, MessageBuilderInteractionComponentEdit, MessageBuilderInteractionFollowupCreate,
    MessageBuilderInteractionFollowupEdit, MessageBuilderInteractionResponseCreate,
    MessageBuilderInteractionResponseEdit
)

from .conversions import CONVERSION_ABORT, CONVERSION_INTERACTION_EVENT, CONVERSION_MESSAGE
from .response_modifier import (
    get_show_for_invoking_user_only_from, get_show_for_invoking_user_only_of, get_wait_for_acknowledgement_of
)


INTERACTION_TYPE_APPLICATION_COMMAND = InteractionType.application_command
INTERACTION_TYPE_MESSAGE_COMPONENT = InteractionType.message_component
INTERACTION_TYPE_FORM_SUBMIT = InteractionType.form_submit
INTERACTION_TYPE_AUTOCOMPLETE = InteractionType.application_command_autocomplete


async def get_request_coroutines(client, interaction_event, response_modifier, response, is_return):
    """
    Gets request coroutine after an output from a command coroutine. Might return `None` if there is nothing to send.
    
    This function is an async iterable coroutine generator.
    
    Parameters
    ----------
    client : ``Client``
        The client who will send the responses if applicable.
    interaction_event : ``InteractionEvent``
        The respective event to respond on.
    response_modifier : `None`, ``ResponseModifier``
        Modifies values returned and yielded to command coroutine processor.
    response : `object`
        Object yielded or returned by the command coroutine.
    is_return : `bool`
        Whether the response is used in a return and we do not require response message.
    
    Yields
    -------
    request_coroutine : `None`, `CoroutineType`
    """
    interaction_event_type = interaction_event.type
    
    # unanswered auto completions are a special case, where we want to auto complete if the event is unanswered.
    if (interaction_event_type is INTERACTION_TYPE_AUTOCOMPLETE) and interaction_event.is_unanswered():
        yield client.interaction_application_command_autocomplete(interaction_event, response)
        return
    
    if (response is None):
        if interaction_event.is_unanswered():
            if (
                (
                    (interaction_event_type is INTERACTION_TYPE_APPLICATION_COMMAND) and
                    (interaction_event.target_type is not ApplicationCommandTargetType.embedded_activity_launch)
                ) or
                (interaction_event_type is INTERACTION_TYPE_FORM_SUBMIT)
            ):
                if (not is_return):
                    # This usually means, the command is uncompleted
                    yield client.interaction_application_command_acknowledge(
                        interaction_event,
                        get_wait_for_acknowledgement_of(response_modifier),
                        show_for_invoking_user_only = get_show_for_invoking_user_only_of(response_modifier),
                    )
                
                return
            
            if interaction_event_type is INTERACTION_TYPE_MESSAGE_COMPONENT:
                if is_return:
                    # Maybe the user launched up an other task to handle the component interaction. If this happened,
                    # we just need to skip a ready cycle
                    await skip_ready_cycle()
                    
                    # If the user indeed launched up an other task, we leave.
                    if not interaction_event.is_unanswered():
                        return
                
                yield client.interaction_component_acknowledge(
                    interaction_event,
                    get_wait_for_acknowledgement_of(response_modifier),
                )
                
                return
            
            if (
                (interaction_event_type is INTERACTION_TYPE_APPLICATION_COMMAND) and
                (interaction_event.target_type is ApplicationCommandTargetType.embedded_activity_launch)
            ):
                if not interaction_event.is_unanswered():
                    return
                
                yield client.interaction_embedded_activity_launch(
                    interaction_event,
                    get_wait_for_acknowledgement_of(response_modifier),
                )
                return
        
        # no more cases
        return
    
    # wait for async acknowledgement if applicable
    await interaction_event._wait_for_async_task_completion()
    
    if isinstance(response, InteractionForm):
        yield client.interaction_form_send(interaction_event, response)
        return
    
    if is_coroutine_generator(response):
        response = await process_command_coroutine_generator(
            client,
            interaction_event,
            response_modifier,
            response,
        )
        
        async for request_coroutine in get_request_coroutines(
            client,
            interaction_event,
            response_modifier,
            response,
            is_return,
        ):
            yield request_coroutine
        
        return
    
    if not isinstance(response, InteractionResponse):
        if isinstance(response, tuple):
            response = InteractionResponse(*response)
        else:
            response = InteractionResponse(response)
    
    for request_coroutine in response.get_request_coroutines(
        client,
        interaction_event,
        response_modifier,
    ):
        yield request_coroutine
    
    # No more cases
    return


async def process_command_coroutine_generator(
    client,
    interaction_event,
    response_modifier,
    coroutine_generator,
):
    """
    Processes a slash command coroutine generator.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The client who will send the responses if applicable.
    interaction_event : ``InteractionEvent``
        The respective event to respond on.
    response_modifier : `None`, ``ResponseModifier``
        Modifies values returned and yielded to command coroutine processor.
    coroutine_generator : `CoroutineGeneratorType`
        A coroutine generator with will send command response.
    
    Returns
    -------
    response : `object`
        Returned object by the coroutine generator.
    
    Raises
    ------
    BaseException
        object exception raised by `coroutine_generator`.
    """
    response_message = None
    response_exception = None
    
    while True:
        if response_exception is None:
            step = coroutine_generator.asend(response_message)
        else:
            step = coroutine_generator.athrow(response_exception)
        
        try:
            response = await step
        except StopAsyncIteration as err:
            # catch `StopAsyncIteration` only if it is a new one.
            if (response_exception is not None) and (response_exception is not err):
                raise
            
            args = err.args
            if args:
                response = args[0]
            else:
                response = None
            break
        
        except InteractionAbortedError as err:
            response = err.response
            break
        
        except BaseException as err:
            if (response_exception is None) or (response_exception is not err):
                raise
            
            raise
        
        else:
            # We set it first, since if `get_request_coroutines` yields nothing, we would be meowed up.
            response_message = None
            response_exception = None
            
            async for request_coroutine in get_request_coroutines(
                client,
                interaction_event,
                response_modifier,
                response,
                False,
            ):
                try:
                    maybe_response_message = await request_coroutine
                except GeneratorExit:
                    raise
                
                except BaseException as err:
                    # `response_message` may have be set before with an iteration, so reset it.
                    response_message = None
                    response_exception = err
                    break
                
                else:
                    # It can happen that the first request creates the response message and the second one edits it.
                    if (maybe_response_message is not None):
                        response_message = maybe_response_message
                        maybe_response_message = None
    return response


async def process_command_coroutine(client, interaction_event, response_modifier, coroutine):
    """
    Processes a slasher application command coroutine.
    
    If the coroutine returns or yields a string or an embed like then sends it to the respective channel.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The client who will send the responses if applicable.
    interaction_event : ``InteractionEvent``
        The respective event to respond on.
    response_modifier : `None`, ``ResponseModifier``
        Modifies values returned and yielded to command coroutine processor.
    coroutine : `Coroutine`
        A coroutine which will produce command response.
    
    Raises
    ------
    BaseException
        object exception raised by `CoroutineType`.
    """
    if is_coroutine_generator(coroutine):
        response = await process_command_coroutine_generator(
            client,
            interaction_event,
            response_modifier,
            coroutine,
        )
    else:
        try:
            response = await coroutine
        except InteractionAbortedError as err:
            response = err.response
    
    async for request_coroutine in get_request_coroutines(
        client,
        interaction_event,
        response_modifier,
        response,
        True,
    ):
        await request_coroutine


class InteractionResponse(
    MessageBuilderInteractionComponentEdit,
    MessageBuilderInteractionFollowupCreate,
    MessageBuilderInteractionFollowupEdit,
    MessageBuilderInteractionResponseCreate,
    MessageBuilderInteractionResponseEdit,
    MessageBuilderBase,
):
    """
    Message builder for interaction responses.
    
    May pass it's parameters to any of the following method depending on control flow.
    
    - ``Client.interaction_application_command_acknowledge``
    - ``Client.interaction_application_command_autocomplete``
    - ``Client.interaction_component_acknowledge``
    - ``Client.interaction_component_message_edit``
    - ``Client.interaction_embedded_activity_launch``
    - ``Client.interaction_followup_message_create``
    - ``Client.interaction_followup_message_edit``
    - ``Client.interaction_response_message_create``
    - ``Client.interaction_response_message_edit``
    
    Attributes
    ----------
    fields : `dict<Conversion, object>`
        The fields to create the message with.
    """
    __slots__ = ()
    
    
    abort = CONVERSION_ABORT
    interaction_event = CONVERSION_INTERACTION_EVENT
    message = CONVERSION_MESSAGE
    
    
    def __new__(cls, *positional_parameters, **keyword_parameters):
        """
        Parameters
        ----------
        *positional_parameters : Positional parameters
            Additional parameters to create the message with.
        
        **keyword_parameters : Keyword parameters
            Additional parameters to create the message with.
        
        Other Parameters
        ----------------
        abort : `None`, `bool`, Optional (Keyword only)
            Whether the interaction response is dropped from ``abort``.
        
        allowed_mentions : `None`,  ``AllowedMentionProxy``, `str`, ``UserBase``, ``Role``, `list` of \
                (`str`, ``UserBase``, ``Role`` ) , Optional
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions``
            for details.
        
        attachments : `None | object`, Optional (Keyword only)
            Attachments to send.
        
        components : `None`, ``Component``, `(tuple | list)<Component, (tuple | list)<Component>>`
            Components attached to the message.
        
        content : `None`, `str`, Optional
            The message's content if given.
        
        poll : `None`, ``Poll``, Optional
            The message's poll.
            
            > Response message must be created or else discord will ignore the `poll` field.
        
        embed : `None`, `Embed`, Optional
            Alternative for `embeds`.
        
        embeds : `None`, `list<Embed>`, Optional
            The new embedded content of the message.
        
        event : ``None | InteractionEvent`` = `None`, Optional
            Alternative for `interaction_event`.
        
        file : `None | object`, Optional (Keyword only)
            Alternative for `attachments`.
        
        files : `None | object`, Optional (Keyword only)
            Alternative for `attachments`.
        
        flags : `int`, ``MessageFlag`, Optional
            The message's flags.
        
        interaction_event : ``None | InteractionEvent`` = `None`, Optional
            A specific event ot answer instead of the command's.
        
        message : ``None | Message``, Optional (Keyword only)
            Whether the interaction's message should be edited.
        
        show_for_invoking_user_only : `bool` = `False`, Optional (Keyword only)
            Whether the sent message should only be shown to the invoking user.
        
        silent : `bool` = `False`, Optional (Keyword only)
            Whether the message should be delivered silently.
        
        suppress_embeds : `bool` = `False`, Optional (Keyword only)
            Whether the message's embeds should be suppressed initially.
        
        tts : `bool` = `False`, Optional (Keyword only)
            Whether the message is text-to-speech.
        
        voice_attachment : ``None | VoiceAttachment``, Optional (Keyword only)
            Modifies the message to be a voice message, allowing it to contain just a single voice attachment.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - if a parameter's value is incorrect.
        """
        positional_parameters_length = len(positional_parameters)
        if not positional_parameters_length:
            self = MessageBuilderBase.__new__(cls)
        
        else:
            self = positional_parameters[0]
            if isinstance(self, cls):
                if positional_parameters_length > 1:
                    self._with_positional_parameters(islice(positional_parameters, 1, positional_parameters_length))
            
            else:
                self = MessageBuilderBase.__new__(cls)
                self._with_positional_parameters(positional_parameters)
        
        if keyword_parameters:
            self._with_keyword_parameters(keyword_parameters)
        
        return self
    
    
    def get_request_coroutines(self, client, interaction_event, response_modifier):
        """
        Gets request coroutine buildable from the ``InteractionResponse``.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        client : ``Client``
            The client who will send the responses if applicable.
        interaction_event : ``InteractionEvent``
            The respective event to respond on.
        response_modifier : `None`, ``ResponseModifier``
            Modifies values returned and yielded to command coroutine processor.
        
        Yields
        -------
        request_coroutine : `None`, `CoroutineType`
        """
        event = self.interaction_event
        if (event is not None):
            interaction_event = event
        
        interaction_event_type = interaction_event.type
        if (
            (interaction_event_type is INTERACTION_TYPE_APPLICATION_COMMAND) or
            (
                (interaction_event_type is INTERACTION_TYPE_FORM_SUBMIT) and
                (interaction_event.message is None)
            )
        ):
            for message in self._try_pull_field_value(CONVERSION_MESSAGE):
                if (response_modifier is not None):
                    response_modifier.apply_to_edition(self)
                
                if message is None:
                    yield client.interaction_response_message_edit(interaction_event, self)
                else:
                    # Note: here we cannot put `poll` on the message.
                    yield client.interaction_followup_message_edit(interaction_event, message, self)
                return
            
            # If an application command interaction with embedded activity launch check whether we want to indeed
            # launch it
            if (
                (interaction_event.target_type is ApplicationCommandTargetType.embedded_activity_launch) and
                interaction_event.is_unanswered() and
                (not self.abort)
            ):
                yield client.interaction_embedded_activity_launch(
                    interaction_event,
                    get_wait_for_acknowledgement_of(response_modifier),
                )
                if (response_modifier is not None):
                    response_modifier.apply_to_creation(self)
                
                yield client.interaction_followup_message_create(interaction_event, self)
                return
            
            if (not interaction_event.is_unanswered()):
                need_acknowledging = False
            elif (self.attachments is not None) or (self.voice_attachment is not None):
                need_acknowledging = True
            elif self.abort:
                need_acknowledging = False
            else:
                need_acknowledging = False
            
            if need_acknowledging:
                yield client.interaction_application_command_acknowledge(
                    interaction_event,
                    show_for_invoking_user_only = get_show_for_invoking_user_only_from(self, response_modifier),
                )
            
            if (response_modifier is not None):
                response_modifier.apply_to_creation(self)
            
            if need_acknowledging or (not interaction_event.is_unanswered()):
                if (interaction_event.message is None) and (not interaction_event.is_responded()):
                    yield client.interaction_response_message_edit(interaction_event, self)
                else:
                    yield client.interaction_followup_message_create(interaction_event, self)
            else:
                yield client.interaction_response_message_create(interaction_event, self)
            
            return
        
        if (
            (interaction_event_type is INTERACTION_TYPE_MESSAGE_COMPONENT) or
            (
                (interaction_event_type is INTERACTION_TYPE_FORM_SUBMIT) and
                (interaction_event.message is not None)
            )
        ):
            if self.abort:
                # If we are aborting we acknowledge it (if not yet) and create a new message.
                if interaction_event.is_unanswered():
                    yield client.interaction_component_acknowledge(interaction_event)
                
                if (response_modifier is not None):
                    response_modifier.apply_to_creation(self)
                
                yield client.interaction_followup_message_create(interaction_event, self)
                
            elif interaction_event.is_unanswered():
                if (self.attachments is not None) or (self.voice_attachment is not None):
                    need_acknowledging = True
                else:
                    need_acknowledging = False
                
                if need_acknowledging:
                    yield client.interaction_component_acknowledge(
                        interaction_event,
                    )
                
                # Note: here we cannot put `poll` on the message.
                if (response_modifier is not None):
                    response_modifier.apply_to_edition(self)
                
                if need_acknowledging:
                    yield client.interaction_component_acknowledge(
                        interaction_event,
                        get_wait_for_acknowledgement_of(response_modifier),
                    )
                
                yield client.interaction_component_message_edit(interaction_event, self)
            
            elif interaction_event.is_deferred():
                yield client.interaction_response_message_edit(interaction_event, self)
            
            elif interaction_event.is_responded():
                if (response_modifier is not None):
                    response_modifier.apply_to_creation(self)
                
                yield client.interaction_followup_message_create(interaction_event, self)
            
            return
        
        if (interaction_event_type is INTERACTION_TYPE_AUTOCOMPLETE):
            if interaction_event.is_unanswered():
                yield client.interaction_application_command_autocomplete(interaction_event, None)
            
            if (response_modifier is not None):
                response_modifier.apply_to_creation(self)
            
            yield client.interaction_followup_message_create(interaction_event, self)
            return
        
        # no more cases


def abort(
    *positional_parameters,
    show_for_invoking_user_only = True,
    **keyword_parameters,
):
    """
    Aborts the slash response with sending the passed parameters as a response.
    
    The abortion auto detects `show_for_invoking_user_only` if not given. Not follows the command's preference.
    If only a string `content` is given, `show_for_invoking_user_only` will become `True`, else `False`. The reason of
    becoming `False` at that case is, Discord ignores every other field except string content.
    
    Parameters
    ----------
    *positional_parameters : Positional parameters
        Additional parameters to create the message with.
    
    show_for_invoking_user_only : `bool` = `True`, Optional (Keyword only)
        Whether the sent message should only be shown to the invoking user.
    
    **keyword_parameters : Keyword parameters
        Additional parameters to create the message with.
    
    Other Parameters
    ----------------
    allowed_mentions : `None`,  ``AllowedMentionProxy``, `str`, ``UserBase``, ``Role``, `list` of \
            (`str`, ``UserBase``, ``Role`` ) , Optional
        Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions``
        for details.
    
    attachments : `None | object`, Optional (Keyword only)
        Attachments to send.
    
    components : `None`, ``Component``, `(tuple | list)<Component, (tuple | list)<Component>>`
        Components attached to the message.
    
    content : `None`, `str`, Optional
        The message's content if given.
    
    poll : `None`, ``Poll``, Optional
        The message's poll.
        
        > Response message must be created or else discord will ignore the `poll` field.
    
    embed : `None`, `Embed`, Optional
        Alternative for `embeds`.
    
    embeds : `None`, `list<Embed>`, Optional
        The new embedded content of the message.
    
    event : ``None | InteractionEvent`` = `None`, Optional
        Alternative for `interaction_event`.
    
    file : `None | object`, Optional (Keyword only)
        Alternative for `attachments`.
    
    files : `None | object`, Optional (Keyword only)
        Alternative for `attachments`.
    
    flags : `int`, ``MessageFlag`, Optional
        The message's flags.
    
    interaction_event : ``None | InteractionEvent`` = `None`, Optional
        A specific event ot answer instead of the command's.
    
    message : ``None | Message``, Optional (Keyword only)
        Whether the interaction's message should be edited.
    
    silent : `bool` = `False`, Optional (Keyword only)
        Whether the message should be delivered silently.
    
    suppress_embeds : `bool` = `False`, Optional (Keyword only)
        Whether the message's embeds should be suppressed initially.
    
    tts : `bool` = `False`, Optional (Keyword only)
        Whether the message is text-to-speech.
    
    voice_attachment : ``None | VoiceAttachment``, Optional (Keyword only)
        Modifies the message to be a voice message, allowing it to contain just a single voice attachment.
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - if a parameter's value is incorrect.
    InteractionAbortedError
        - The exception which aborts the interaction, then yields the response.
    """
    raise InteractionAbortedError(
        InteractionResponse(
            *positional_parameters,
            **keyword_parameters,
            abort = True,
            show_for_invoking_user_only = show_for_invoking_user_only,
        )
    )


class InteractionAbortedError(BaseException):
    """
    An ``InteractionAbortedError`` is raised when a slash command is aborted. This class holds the response to send to
    the client.
    
    Attributes
    ----------
    response : ``InteractionResponse``
        The response to send.
    """
    __slots__ = ('response',)
    
    # This is to support keyword parameters
    __init__ = object.__init__
    
    def __new__(cls, response):
        """
        Creates a new ``InteractionAbortedError`` with the given response.
        
        Parameters
        ----------
        response : ``InteractionResponse``
            The response to send.
        """
        self = BaseException.__new__(cls, response)
        self.response = response
        return self
    
    
    def __repr__(self):
        """Returns the exception's representation."""
        return f'{self.__class__.__name__}({self.response!r})'
    
    
    def __eq__(self, other):
        """Returns whether the two abortions are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.response != other.response:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the abortion exception hash value"""
        return hash(self.response)



__all__ = ()

from codecs import decode

from scarletio import RichAttributeErrorBaseType


ESCAPE_MAP = {
    '\\':'\\',
    '\'': '\'',
    '"': '"',
    'a': '\a',
    'b': '\b',
    'f': '\f',
    'n': '\n',
    'r': '\r',
    't': '\t',
    'v': '\v',
}


ESCAPE_MAP_REVERSED = {value: key for key, value in ESCAPE_MAP.items()}


PARSED_STATE_SUCCESS = 1 << 0
PARSED_STATE_FAILURE = 1 << 1
PARSED_STATE_END = 1 << 2

ERROR_CODE_VALUE_ALREADY_PARSED = 1
ERROR_CODE_EXPECTED_EQUAL_SIGN = 2
ERROR_CODE_QUOTE_NOT_CLOSED = 3
ERROR_CODE_KEY_STARTER_INVALID = 4


ERROR_MESSAGES = {
    ERROR_CODE_VALUE_ALREADY_PARSED: (
        'Expected comment or end of line.\n'
        'Perhaps there are extra character(s) after the value?'
    ),
    ERROR_CODE_EXPECTED_EQUAL_SIGN: (
        'Expected equal sign, comment or end of line.\n'
        'Perhaps there are extra character(s) after the key?'
    ),
    ERROR_CODE_QUOTE_NOT_CLOSED: (
        'Expected the quote to be closed.\n'
        'If there is no quote on this line at all check them above.'
    ),
    ERROR_CODE_KEY_STARTER_INVALID: (
        'Expected alphabetical character or underscore.\n'
        'Keys cannot start with numbers, perhaps it started with one?'
    ),
}


def embed_error_code(parsed_state, error_code):
    """
    Embeds the error code into the given parsed state.
    
    Parameters
    ----------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    error_code : `int`
        The error code to embed.
    
    Returns
    -------
    parsed_state : `int`
    """
    return (error_code << 16) | parsed_state


def uproot_error_code(parsed_state):
    """
    Uproots the error code from the given parsed state.
    
    Parameters
    ----------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    
    Returns
    -------
    error_code : `int`
    """
    return (parsed_state >> 16) & 0xffff


class ParserState(RichAttributeErrorBaseType):
    """
    Represents parser's state.
    
    Attributes
    ----------
    end : `int`
        The end position of the parser.
    line_index : `int`
        The index of the current line.
    line_start : `int`
        Where the current line is started.
    position : `int`
        The current position.
    value : `list<str>`
        The value to parse by character.
    """
    __slots__ = ('end', 'line_index', 'line_start', 'position', 'value')
    
    def __new__(cls, value):
        """
        Parameters
        ----------
        value : `str`
            The value to parse.
        """
        self = object.__new__(cls)
        self.end = len(value)
        self.line_index = 0
        self.line_start = 0
        self.position = 0
        self.value = [*value] # When using C change it to uint32 array.
        return self
    
    
    def __repr__(self):
        """Returns the parser's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' position = ')
        repr_parts.append(repr(self.position))
        repr_parts.append(' / ')
        repr_parts.append(repr(self.end))
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def get_value_in_range(self, start, end):
        """
        Returns the value of the parser state at the given range.
        
        Parameters
        ----------
        start : `int`
            Start index.
        end : `int`
            End index.
        
        Returns
        -------
        value_in_range : `str`
        """
        return ''.join(self.value[start : end])
    
        
    def set_line_break(self):
        """
        Sets the line last break of the parser to the current index.
        """
        self.line_start = self.position
        self.line_index += 1


def render_escaped_string_into(into, string):
    """
    Renders the string as escaped into the given container.
    
    Parameters
    ----------
    into : `list<str>`
        String parts to extend.
    string : `str`
        The string to render.
    
    Returns
    -------
    into : `list<str>`
    """
    for character in string:
        try:
            character = ESCAPE_MAP_REVERSED[character]
        except KeyError:
            pass
        else:
            into.append('\\')
        
        into.append(character)
    
    return into


def render_string_hidden_representation_into(into, string, cut_at):
    """
    Renders the string as escaped into the given container.
    
    Parameters
    ----------
    into : `list<str>`
        String parts to extend.
    string : `str`
        The string to render.
    cut_at : `int`
        Where to cut the string.
    
    Returns
    -------
    into : `list<str>`
    """
    string_length = len(string)
    
    into.append('\'')
    
    # Case: '' (so 0 characters)
    if string_length == 0:
        pass
    
    # Case: 'a'  to 'aaabbb' (1 to  1 - cut_at * 2 characters)
    elif string_length <= cut_at << 1:
        render_escaped_string_into(into, string)
    
    # Case: 'aaa ... (+n hidden) ... bbb' (so cut_at * 2 + 1 to n characters)
    else:
        render_escaped_string_into(into, string[:cut_at])
        into.append(' ... (+')
        into.append(repr(string_length - (cut_at << 1)))
        into.append(' hidden) ... ')
        render_escaped_string_into(into, string[-cut_at:])
    
    into.append('\'')

    return into


class ParserFailureInfo(RichAttributeErrorBaseType):
    """
    Holds additional information about parsing failure.
    
    Attributes
    ----------
    index : `int`
        The error position's index on the `line`. (0 based.)
    line : `str`
        The line where the error was found.
    line_index : `int`
        The `line`'s index inside of the file. (0 based.)
    error_code : `int`
        Error representing the reason why parsing failed.
    """
    __slots__ = ('error_code', 'index', 'line', 'line_index')
    
    def __new__(cls, index, line, line_index, error_code):
        """
        Creates a new parser failure info with the given parameters.
        
        Parameters
        ----------
        index : `int`
            The error position's index on the `line`. (0 based.)
        line : `str`
            The line where the error was found.
        line_index : `int`
            The `line`'s index inside of the file. (0 based.)
        error_code : `int`
            Error representing the reason why parsing failed.
        """
        self = object.__new__(cls)
        self.index = index
        self.line = line
        self.line_index = line_index
        self.error_code = error_code
        return self
    
    
    @classmethod
    def from_parser_state(cls, parser_state, error_code):
        """
        Creates a new parser failure info from the given parser state.
        
        Parameters
        ----------
        parser_state : ``ParserState``
            The parser state to get the error information from.
        error_code : `int`
            Error representing the reason why parsing failed.
        
        Returns
        -------
        self : `instance<cls>`
        """
        line_start_position = parser_state.line_start
        index = parser_state.position - line_start_position
        line_index = parser_state.line_index
        
        exhaust_line(parser_state)
        line = parser_state.get_value_in_range(line_start_position, parser_state.position).rstrip()
        return cls(index, line, line_index, error_code)
    
    
    def __repr__(self):
        """Returns the parser failure's representation."""
        repr_parts = ['<', self.__class__.__name__]
        repr_parts.append(' line_index = ')
        repr_parts.append(repr(self.line_index))
        repr_parts.append(', index = ')
        repr_parts.append(repr(self.index))
        repr_parts.append(', line = ')
        render_string_hidden_representation_into(repr_parts, self.line, 3)
        repr_parts.append(', error_code = ')
        repr_parts.append(repr(self.error_code))
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two parser failure infos are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.error_code != other.error_code:
            return False
        
        if self.index != other.index:
            return False
        
        if self.line != other.line:
            return False
        
        if self.line_index != other.line_index:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the parser failure handler."""
        hash_value = 0
        
        # error_code
        hash_value ^= self.error_code
        
        # index
        hash_value ^= self.index << 8
        
        # line
        hash_value ^= hash(self.line)
        
        # line_index
        hash_value ^= self.line_index << 24
        
        return hash_value
    
    
    def get_error_message(self):
        """
        Gets the error message attached to the failure info.
        
        Returns
        -------
        error_message : `None | str`
        """
        return ERROR_MESSAGES.get(self.error_code, None)


def is_character_white_space(character):
    """
    Returns whether the character is a white space.
    
    Parameters
    ----------
    character : `str`
        The character to check.
    
    Returns
    -------
    is_white_space : `bool`
    """
    return character in (' ', '\t')


def is_character_per_r(character):
    """
    Returns whether the character is a `\\r`.
    
    Parameters
    ----------
    character : `str`
        The character to check.
    
    Returns
    -------
    is_per_r : `bool`
    """
    return character == '\r'


def is_character_per_n(character):
    """
    Returns whether the character is a `\\r`.
    
    Parameters
    ----------
    character : `str`
        The character to check.
    
    Returns
    -------
    is_per_n : `bool`
    """
    return character == '\n'


def is_character_comment(character):
    """
    Returns whether the character is a comment sign.
    
    Parameters
    ----------
    character : `str`
        The character to check.
    
    Returns
    -------
    is_comment : `bool`
    """
    return character == '#'


def is_character_non_comment_or_line_break(character):
    """
    Returns whether the character is NOT a comment sign or line break.
    
    Parameters
    ----------
    character : `str`
        The character to check.
    
    Returns
    -------
    is_comment : `bool`
    """
    return character not in ('#', '\n', '\r')


def is_character_equal(character):
    """
    Returns whether the character is an equal sign.
    
    Parameters
    ----------
    character : `str`
        The character to check.
    
    Returns
    -------
    is_equal : `bool`
    """
    return character == '='


def is_character_identifier_starter(character):
    """
    Returns whether the character is an identifier starter.
    
    Parameters
    ----------
    character : `str`
        The character to check.
    
    Returns
    -------
    is_character_identifier_starter : `bool`
    """
    return character.isalpha() or character == '_'


def is_character_identifier_continuous(character):
    """
    Returns whether the character is an identifier continuous.
    
    Parameters
    ----------
    character : `str`
        The character to check.
    
    Returns
    -------
    is_character_identifier_starter : `bool`
    """
    return character.isalpha() or character == '_' or character.isnumeric()


def is_at_end(parser_state):
    """
    Returns whether the parser state is at the end position.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to check its position.
    
    Returns
    -------
    is_at_end : `bool`
    """
    return parser_state.position >= parser_state.end


def exhaust_any(parser_state):
    """
    Exhausts any characters.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    """
    if is_at_end(parser_state):
        return PARSED_STATE_END
    
    parser_state.position += 1
    return PARSED_STATE_SUCCESS


def exhaust_if(parser_state, condition):
    """
    Exhausts the next character if the given condition passes on it.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    condition : `(str) -> bool`
        Condition to call on the next character.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    
    """
    if is_at_end(parser_state):
        return PARSED_STATE_END
    
    parser_state_position = parser_state.position
    character_at_position = parser_state.value[parser_state_position]
    if not condition(character_at_position):
        return PARSED_STATE_FAILURE
    
    parser_state.position = parser_state_position + 1
    return PARSED_STATE_SUCCESS


def exhaust_white_space(parser_state):
    """
    Exhausts the next white space character.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    """
    return exhaust_if(parser_state, is_character_white_space)


def exhaust_line_break(parser_state):
    """
    Exhausts the next line break. It can be either `\r\n`, `\n`, `\r`.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    """
    parsed_state = exhaust_if(parser_state, is_character_per_r)
    if parsed_state & PARSED_STATE_END:
        return parsed_state
    
    if parsed_state & PARSED_STATE_SUCCESS:
        if exhaust_if(parser_state, is_character_per_n) & PARSED_STATE_END:
            parsed_state |= PARSED_STATE_END
        
        parser_state.set_line_break()
        return parsed_state
    
    parsed_state = exhaust_if(parser_state, is_character_per_n)
    if parsed_state & PARSED_STATE_SUCCESS:
        parser_state.set_line_break()
    
    return parsed_state


def exhaust_non_comment_or_line_break(parser_state):
    """
    Exhausts the next non comment or line break character.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    """
    return exhaust_if(parser_state, is_character_non_comment_or_line_break)


def repeat_exhauster(parser_state, exhauster):
    """
    Repeats the given exhauster.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    exhauster : `(ParserState) -> int`
        Exhauster to repeat.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    """
    return_parsed_state = 0
    
    while True:
        parsed_state = exhauster(parser_state)
        # A parsing can end with `end | success` and `failure`
        if parsed_state & PARSED_STATE_SUCCESS:
            return_parsed_state |= PARSED_STATE_SUCCESS
        
        if parsed_state & PARSED_STATE_END:
            return_parsed_state |= PARSED_STATE_END
            return return_parsed_state
        
        if parsed_state & PARSED_STATE_FAILURE:
            break
        
        continue
    
    if return_parsed_state:
        return return_parsed_state
    
    return PARSED_STATE_FAILURE


def exhaust_white_space_all(parser_state):
    """
    Exhausts all white space in a row.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    """
    return repeat_exhauster(parser_state, exhaust_white_space)


def exhaust_line_break_all(parser_state):
    """
    Exhausts all line breaks in a row.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    """
    return repeat_exhauster(parser_state, exhaust_line_break)


def chain_exhausters(parser_state, exhausters):
    """
    Chains the given exhausters after each other and repeats them while any of them passes.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    exhausters : `tuple<(ParserState) -> int>`
        Exhauster to repeat.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    """
    return_parsed_state = 0
    
    while True:
        for exhauster in exhausters:
            parsed_state = exhauster(parser_state)
            if parsed_state & PARSED_STATE_SUCCESS:
                return_parsed_state |= PARSED_STATE_SUCCESS
            
            if parsed_state & PARSED_STATE_END:
                return_parsed_state |= PARSED_STATE_END
                return return_parsed_state
            
            if parsed_state & PARSED_STATE_FAILURE:
                continue
            
            # If we did not fail, lets start over
            break
        
        else:
            break
        
        continue
    
    if return_parsed_state:
        return return_parsed_state
    
    return PARSED_STATE_FAILURE


def exhaust_line_break_and_white_space_all(parser_state):
    """
    Exhausts all line breaks and white space in a row.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    """
    return chain_exhausters(parser_state, (exhaust_white_space_all, exhaust_line_break_all))


def exhaust_line(parser_state):
    """
    Exhausts the whole line. Ends after the first line break.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    """
    while True:
        parsed_state = exhaust_line_break(parser_state)
        if parsed_state & PARSED_STATE_SUCCESS:
            return parsed_state
        
        if parsed_state & PARSED_STATE_END:
            parsed_state |= PARSED_STATE_SUCCESS
            return parsed_state
        
        exhaust_any(parser_state)
        continue


def exhaust_till_comment_or_line_break(parser_state):
    """
    Exhausts till the next comment or line break.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    """
    return repeat_exhauster(parser_state, exhaust_non_comment_or_line_break)


def exhaust_comment(parser_state):
    """
    Exhausts the commented value.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    """
    parsed_state = exhaust_if(parser_state, is_character_comment)
    if parsed_state & (PARSED_STATE_FAILURE | parsed_state & PARSED_STATE_END):
        return parsed_state
    
    if exhaust_line(parser_state) & PARSED_STATE_END:
        parsed_state |= PARSED_STATE_END
    
    return parsed_state


def parse_key_unquoted(parser_state):
    """
    Parses an unquoted key.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    value : `None | str`
        The parsed out value.
    """
    parser_state_position = parser_state.position
    
    parsed_state = exhaust_if(parser_state, is_character_identifier_starter)
    if not (parsed_state & PARSED_STATE_SUCCESS):
        if parsed_state & PARSED_STATE_FAILURE:
            parsed_state = embed_error_code(parsed_state, ERROR_CODE_KEY_STARTER_INVALID)
        return parsed_state, None
    
    while True:
        continuous_parsed_state = exhaust_if(parser_state, is_character_identifier_continuous)
        if continuous_parsed_state & PARSED_STATE_SUCCESS:
            continue
        
        if continuous_parsed_state & PARSED_STATE_FAILURE:
            break
        
        if continuous_parsed_state & PARSED_STATE_END:
            parsed_state |= PARSED_STATE_END
            break
        
        # No more cases
        continue
    
    value = parser_state.get_value_in_range(parser_state_position, parser_state.position)
    return parsed_state, value


def parse_value_unquoted(parser_state):
    """
    Parses an unquoted value.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    value : `None | str`
        The parsed out value.
    """
    parser_state_position = parser_state.position
    parsed_state = exhaust_till_comment_or_line_break(parser_state)
    if parsed_state & (PARSED_STATE_SUCCESS | PARSED_STATE_END):
        value = parser_state.get_value_in_range(parser_state_position, parser_state.position).rstrip()
        if not value:
            value = None
    else:
        value = None
    
    return parsed_state, value


def parse_next(parser_state):
    """
    Parses the next character.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    value : `None | str`
        The parsed out value.
    """
    if is_at_end(parser_state):
        return PARSED_STATE_END, None
    
    parser_state_position = parser_state.position
    value = parser_state.value[parser_state_position]
    parser_state.position = parser_state_position + 1
    return PARSED_STATE_SUCCESS, value


def parse_quoted_content(parser_state, expected_ending):
    """
    Parses a quoted value's (or key's) content till it ends with the expected ending.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    expected_ending : `str`
        The expected ending character of the string.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    value : `None | str`
        The parsed out value.
    """
    value_parts = []
    
    while True:
        parsed_state, value = parse_next(parser_state)
        if parsed_state == PARSED_STATE_END:
            parsed_state = embed_error_code(parsed_state | PARSED_STATE_FAILURE, ERROR_CODE_QUOTE_NOT_CLOSED)
            break
        
        if value == expected_ending:
            break
        
        if value != '\\':
            value_parts.append(value)
            continue
        
        parsed_state, value = parse_next(parser_state)
        if parsed_state == PARSED_STATE_END:
            value_parts.append(value)
            parsed_state = embed_error_code(parsed_state | PARSED_STATE_FAILURE, ERROR_CODE_QUOTE_NOT_CLOSED)
            break
        
        escaped = ESCAPE_MAP.get(value, None)
        if escaped is None:
            value_parts.append('\\')
            value_parts.append(value)
        else:
            value_parts.append(escaped)
    
    if value_parts:
        value = decode(''.join(value_parts), 'unicode-escape')
    else:
        value = None
    
    return parsed_state, value


def parse_key_or_value(parser_state, parsing_key):
    """
    Parses a key or a value.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    parsing_key : `bool`
        Whether we are parsing a key.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    value : `None | str`
        The parsed out value.
    """
    # Should not happen -> do not embed error message
    if is_at_end(parser_state):
        return PARSED_STATE_END | PARSED_STATE_FAILURE, None
    
    parser_state_position = parser_state.position
    character_at_position = parser_state.value[parser_state_position]
    if character_at_position in ('\'', '"'):
        parser_state.position = parser_state_position + 1
        return parse_quoted_content(parser_state, character_at_position)
    
    if parsing_key:
        parser = parse_key_unquoted
    else:
        parser = parse_value_unquoted
    
    return parser(parser_state)


def maybe_build_item(key, value):
    """
    Builds a parsed item.
    
    Parameters
    ----------
    key : `None | str`
        Parsed key.
    value : `None | str`
        The parsed out value.
    
    Returns
    -------
    item : `None`, `tuple` (`str`, `str` | `None`)
    """
    if (key is not None):
        return (key, value)


def parse_item_part_end(parser_state, key, value):
    """
    Parses an item's part's end.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to exhaust characters from.
    key : `None | str`
        Already parsed out key.
    value : `None | str`
        Already parsed out value.
    
    Returns
    -------
    parse_state_item_pair : `None` | `tuple` (`int`, `None`, `tuple` (`str`, `str` | `None`))
        If the item ended returns a `tuple` of `parsed_state` and `item` fields.
    """
    # white space
    parsed_state = exhaust_white_space_all(parser_state)
    if parsed_state & PARSED_STATE_END:
        return parsed_state | PARSED_STATE_SUCCESS, maybe_build_item(key, value)
    
    # line break
    parsed_state = exhaust_line_break(parser_state)
    if parsed_state & (PARSED_STATE_END | PARSED_STATE_SUCCESS):
        return parsed_state | PARSED_STATE_SUCCESS, maybe_build_item(key, value)
    
    # comment sign ?
    parsed_state = exhaust_comment(parser_state)
    if parsed_state & (PARSED_STATE_END | PARSED_STATE_SUCCESS):
        return parsed_state | PARSED_STATE_SUCCESS, maybe_build_item(key, value)
    
    return None


def parse_item(parser_state):
    """
    Parses out an item from the parser.
    
    Parameters
    ----------
    parser_state : ``ParserState``
        The parser state to parse the next item from.
    
    Returns
    -------
    parsed_state : `int`
        Bitwise flag containing the state how the operation went.
    item : `None`, `tuple` (`str`, `str` | `None`)
        The parsed out item if any.
    """
    # We parse till there is something
    while True:
        parsed_state = exhaust_line_break_and_white_space_all(parser_state)
        if parsed_state & PARSED_STATE_END:
            return parsed_state | PARSED_STATE_SUCCESS, None
        
        parsed_state = exhaust_comment(parser_state)
        if parsed_state & PARSED_STATE_END:
            return parsed_state | PARSED_STATE_SUCCESS, None
        
        if parsed_state & PARSED_STATE_SUCCESS:
            continue
        
        break
    
    # key = value
    # ^^^
    parsed_state, key = parse_key_or_value(parser_state, True)
    if parsed_state & PARSED_STATE_FAILURE:
        # Ignore value on failure
        return parsed_state, None
    
    if parsed_state & PARSED_STATE_END:
        return parsed_state, maybe_build_item(key, None)
    
    # parse: white space | linebreak | comment
    end_of_line_output = parse_item_part_end(parser_state, key, None)
    if end_of_line_output is not None:
        return end_of_line_output
    
    # key = value
    #     ^
    parsed_state = exhaust_if(parser_state, is_character_equal)
    if parsed_state & (PARSED_STATE_END | PARSED_STATE_FAILURE):
        return embed_error_code(parsed_state, ERROR_CODE_EXPECTED_EQUAL_SIGN), maybe_build_item(key, None)
    
    # parse: white space | linebreak | comment
    end_of_line_output = parse_item_part_end(parser_state, key, None)
    if end_of_line_output is not None:
        return end_of_line_output
    
    # key = value
    #       ^^^^^
    parsed_state, value = parse_key_or_value(parser_state, False)
    if parsed_state & PARSED_STATE_FAILURE:
        # Ignore value on failure
        return parsed_state, maybe_build_item(key, None)
    
    # parse: white space | linebreak | comment
    end_of_line_output = parse_item_part_end(parser_state, key, value)
    if end_of_line_output is not None:
        return end_of_line_output
    
    # The line is not ended?
    return embed_error_code(PARSED_STATE_FAILURE, ERROR_CODE_VALUE_ALREADY_PARSED), maybe_build_item(key, value)


def add_item(variables, item):
    """
    Adds an item to the variables.
    
    Parameters
    ----------
    variables : `dict<str, str>`
        The parsed variables.
    item : `None`, `tuple` (`str`, `None` | `str`)
        The item to add.
    
    Returns
    -------
    success : `bool`
    """
    if item is None:
        return False
    
    key, value = item
    if variables.get(key, None) is not None:
        return False
        
    variables[key] = value
    return True


def parse_variables(content):
    """
    Parses out the variables from the given input content.
    
    Parameters
    ----------
    content : `str`
        The content to parse.
    
    Returns
    -------
    variables : `dict` of (`str`, `str` | `None`) items
        The parsed out variables.
    parser_failure_info : `None`, ``ParserFailureInfo``
        Error position if any. `-1` if not applicable.
    """
    variables = {}
    parser_failure_info = None
    
    parser_state = ParserState(content)
    
    while True:
        parsed_state, item = parse_item(parser_state)
        add_item(variables, item)
        
        if parsed_state & PARSED_STATE_FAILURE:
            parser_failure_info = ParserFailureInfo.from_parser_state(parser_state, uproot_error_code(parsed_state))
            break
        
        if parsed_state & PARSED_STATE_END:
            break
        
        if parsed_state & PARSED_STATE_SUCCESS:
            continue
        
        # No other case
        break
    
    return variables, parser_failure_info



__all__ = ('SolarPlayerBase', )

from math import floor
from time import monotonic

from scarletio import Future, RichAttributeErrorBaseType, Task

from ...discord.channel import Channel, ChannelType, create_partial_channel_from_id
from ...discord.client.request_helpers import get_channel_id
from ...discord.core import GUILDS, KOKORO

from .constants import (
    LAVALINK_BAND_COUNT, LAVALINK_KEY_GUILD_ID, LAVALINK_KEY_NODE_OPERATION, LAVALINK_KEY_NODE_OPERATION_PLAYER_DESTROY,
    LAVALINK_KEY_NODE_OPERATION_PLAYER_FILTER, LAVALINK_KEY_NODE_OPERATION_PLAYER_PAUSE,
    LAVALINK_KEY_NODE_OPERATION_PLAYER_PLAY, LAVALINK_KEY_NODE_OPERATION_PLAYER_SEEK,
    LAVALINK_KEY_NODE_OPERATION_PLAYER_STOP, LAVALINK_KEY_NODE_OPERATION_VOICE_UPDATE, LAVALINK_KEY_NO_REPLACE,
    LAVALINK_KEY_PAUSE, LAVALINK_KEY_POSITION_MS, LAVALINK_KEY_START_TIME
)
from .track import ConfiguredTrack


class SolarPlayerBase(RichAttributeErrorBaseType):
    """
    Base player of solar link.
    
    Attributes
    ----------
    _filters : `dict` of (`int`, ``Filter``) items
        The applied filters to the player if any.
    _forward_data : `None`, `dict<str, object>`
        Json to forward to the player's node as necessary.
    _position : `float`
        The position of the current track.
    _position_update : `float`
        When ``._position`` was last updated in monotonic time.
    channel_id : `int`
        The channel's identifier, where the node is connected to.
    guild_id : `int`
        The guild's identifier, where the player is.
    node : ``SolarNode``
        The node that the player is connected to. Defaults to `None`.
    
    
    When subclassing please overwrite the following methods:
    
    - ``.is_paused``
    - ``.get_current``
    
    """
    __slots__ = ('_filters', '_forward_data', '_position', '_position_update', 'channel_id', 'guild_id', 'node',)
    
    def __new__(cls, node, guild_id, channel_id):
        """
        Creates a new solar player instance.
        
        Parameters
        ----------
        node : ``SolarNode``
            The node of the player.
        guild_id : `int`
            The guild's identifier, where the node will connect to.
        channel_id : `int`
            The channel's identifier, where the node will connect to.
        
        Returns
        -------
        self : ``SolarPlayerBase``
            The player.
        waiter : ``Future``
            A future, which can be awaited to wait till the player is connected.
        """
        self = object.__new__(cls)
        
        self._filters = {}
        self._position = 0.0
        self._position_update = 0.0
        self._forward_data = None
        
        self.guild_id = guild_id
        self.channel_id = channel_id
        self.node = node
        
        waiter = Future(KOKORO)
        Task(KOKORO, self._connect(waiter = waiter))
        
        return self, waiter
    
    
    async def _connect(self, waiter = None):
        """
        Connecting task of the solar player.
        
        This method is a coroutine.
        
        Parameters
        ----------
        waiter : `None`, ``Future``
            Waiter to set it's result or exception, when connection is established.
        """
        try:
            client = self.node.client
            
            gateway = client.gateway_for(self.guild_id)
            await gateway.change_voice_state(self.guild_id, self.channel_id)
        
        except GeneratorExit as err:
            if (waiter is not None):
                waiter.set_exception_if_pending(err)
                waiter = None
            
            raise
        
        except BaseException as err:
            if (waiter is not None):
                waiter.set_exception_if_pending(err)
                waiter = None
        
        if (waiter is not None):
            waiter.set_result_if_pending(self)
            waiter = None
    
    
    def is_playing(self):
        """
        Returns whether the player is currently playing anything.
        
        Returns
        -------
        is_playing : `bool`
        """
        if not self.channel_id:
            is_playing = False
        elif self.is_paused():
            is_playing = False
        elif (self.get_current() is not None):
            is_playing = True
        else:
            is_playing = True
        
        return is_playing
    
    
    def is_paused(self):
        """
        Returns whether the player is currently paused.
        
        Returns
        -------
        is_paused : `bool`
        """
        raise NotImplementedError
    
    
    def is_connected(self):
        """
        Returns whether the player is connected to a voice channel.
        
        Returns
        -------
        is_connected : `bool`
        """
        if self.node is None:
            is_connected = False
        elif self.channel_id:
            is_connected = True
        else:
            is_connected = False
        
        return is_connected
    
    
    @property
    def position(self):
        """
        Returns the position in the track.
        
        Returns
        -------
        position : `float`
        """
        current_track = self.get_current()
        if (current_track is None):
            position = 0.0
        else:
            position = self._position
            
            if not self.is_paused():
                position += monotonic() - self._position_update
            
            duration = current_track.track.duration
            if position > duration:
                position = duration
        
        return position
    
    
    def get_current(self):
        """
        Returns the currently playing or paused track of the player.
        
        Returns
        -------
        track : `None`, ``ConfiguredTrack``
        """
        raise NotImplementedError
    
    
    def add_filter(self, filter):
        """
        Adds a filter to the player.
        
        Parameters
        ----------
        filter : ``Filter``
            The filter to add.
        """
        if filter:
            self._filters[filter.identifier] = filter
        else:
            try:
                del self._filters[filter.identifier]
            except KeyError:
                pass
    
    
    def get_filter(self, filter_type):
        """
        Gets the filter of the given type.
        
        Returns
        -------
        filter : `None`, ``Filter``
        """
        return self._filters.get(filter_type.identifier, None)
    
    
    def iter_filters(self):
        """
        Iterates over the filters of the player.
        
        This method is an iterable generator.
        
        Yields
        ------
        filter : ``Filter``
        """
        yield from self._filters.values()
    
    
    def remove_filter(self, filter_type):
        """
        Removes the filter of the given type and returns it.
        
        Returns
        -------
        filter : `None`, ``Filter``
        """
        return self._filters.pop(filter_type.identifier, None)
    
    
    async def apply_filters(self):
        """
        Applies the filters added to the player.
        
        This method is a coroutine.
        """
        node = self.node
        if (node is not None):
            data = {
                LAVALINK_KEY_NODE_OPERATION: LAVALINK_KEY_NODE_OPERATION_PLAYER_FILTER,
                LAVALINK_KEY_GUILD_ID: str(self.guild_id),
            }
            
            for filter in self._filters.values():
                data[filter.json_key] = filter.to_data()
            
            await node._send(data)
    
    
    async def _play(self, configured_track, replace = True):
        """
        Sends a play payload to the player's node.
        
        This method is a coroutine.
        
        Parameters
        ----------
        configured_track : ``ConfiguredTrack``
            The track to play.
        replace : `bool` = `True`, Optional
            Whether the current operation, like playing and pausing should be overwritten.
        """
        node = self.node
        if (node is not None):
            await node._send({
                LAVALINK_KEY_NODE_OPERATION: LAVALINK_KEY_NODE_OPERATION_PLAYER_PLAY,
                LAVALINK_KEY_GUILD_ID: str(self.guild_id),
                LAVALINK_KEY_NO_REPLACE: (not replace),
                **configured_track.un_pack(),
            })
    
    
    async def _stop(self):
        """
        Sends a stop payload to the player's node.
        
        This method is a coroutine.
        """
        node = self.node
        if (node is not None):
            await node._send({
                LAVALINK_KEY_NODE_OPERATION: LAVALINK_KEY_NODE_OPERATION_PLAYER_STOP,
                LAVALINK_KEY_GUILD_ID: str(self.guild_id),
            })
    
    
    async def _pause(self):
        """
        Sends a pause payload to the player's node.
        
        This method is a coroutine.
        """
        node = self.node
        if (node is not None):
            await node._send({
                LAVALINK_KEY_NODE_OPERATION: LAVALINK_KEY_NODE_OPERATION_PLAYER_PAUSE,
                LAVALINK_KEY_GUILD_ID: str(self.guild_id),
                LAVALINK_KEY_PAUSE: True,
            })
    
    
    async def _resume(self):
        """
        Sends a resume payload to the player's node.
        
        This method is a coroutine.
        """
        node = self.node
        if (node is not None):
            await node._send({
                LAVALINK_KEY_NODE_OPERATION: LAVALINK_KEY_NODE_OPERATION_PLAYER_PAUSE,
                LAVALINK_KEY_GUILD_ID: str(self.guild_id),
                LAVALINK_KEY_PAUSE: False,
            })
    
    
    async def _seek(self, position):
        """
        Sends a seek payload to the player's node.
        
        This method is a coroutine.
        
        Parameters
        ----------
        position : `float`
            The new position in seconds.
        """
        node = self.node
        if (node is not None):
            await node._send({
                LAVALINK_KEY_NODE_OPERATION: LAVALINK_KEY_NODE_OPERATION_PLAYER_SEEK,
                LAVALINK_KEY_GUILD_ID: str(self.guild_id),
                LAVALINK_KEY_POSITION_MS: floor(position * 1000.0),
        })
    
    
    async def _destroy(self):
        """
        Sends a destroy payload to the player's node.
        
        This method is a coroutine.
        """
        node = self.node
        if (node is not None) and node.available:
            await node._send({
                LAVALINK_KEY_NODE_OPERATION: LAVALINK_KEY_NODE_OPERATION_PLAYER_DESTROY,
                LAVALINK_KEY_GUILD_ID: str(self.guild_id),
            })
    
    
    async def _voice_update(self):
        """
        Sends voice update payload to the player's node.
        
        This method is a coroutine.
        """
        node = self.node
        if (node is not None):
            forward_data = self._forward_data
            if (forward_data is not None) and (len(forward_data) == 2):
                await node._send(
                    {
                        LAVALINK_KEY_GUILD_ID: str(self.guild_id),
                        LAVALINK_KEY_NODE_OPERATION: LAVALINK_KEY_NODE_OPERATION_VOICE_UPDATE,
                        **self._forward_data,
                    }
                )
    
    
    def _update_state(self, data):
        """
        Updates the position of the player.
        
        Parameters
        ----------
        data : `dict<str, object>`
            State data given.
        """
        try:
            position = data[LAVALINK_KEY_POSITION_MS]
        except KeyError:
            position = 0.0
        else:
            position = position * 0.001
        
        self._position = position
        self._position_update = monotonic()
        # There is also a time key, but dunno why I would use it.
    
    
    async def disconnect(self):
        """
        Disconnects the player from Discord.
        
        This method is a coroutine.
        """
        node = self.node
        if (node is None):
            return
            
        client = node.client
        
        try:
            del client.solarlink.players[self.guild_id]
        except KeyError:
            pass
        
        await self._destroy()
        
        gateway = client.gateway_for(self.guild_id)
        await gateway.change_voice_state(self.guild_id, 0)
        
        self.node = None
    
    
    async def move_to(self, channel):
        """
        Move the voice client to an another voice channel.
        
        This method is a coroutine.
        
        Parameters
        ---------
        channel : ``Channel``, `int`
            The channel where the voice client will move to.
        
        Returns
        -------
        moved : `bool`
            Returns `False` if the voice client is already in the channel or if the voice client is already
            disconnected (should not happen).
        
        Raises
        ------
        TypeError
            If  `channel` was not given as ``Channel`` not `int`.
        """
        node = self.node
        if node is None:
            # Already disconnected.
            return False
        
        channel_id = get_channel_id(channel, Channel.is_in_group_guild_connectable)
        
        if self.channel_id == channel_id:
            return False
        
        gateway = node.client.gateway_for(self.guild_id)
        await gateway.change_voice_state(self.guild_id, channel_id)
        return True
    
    
    async def change_node(self, node):
        """
        Changes the player's node
        
        Parameters
        ----------
        node ``SolarNode``
            The node the player is changed to.
        """
        await self._destroy()
        
        self.node = node
        
        await self._voice_update()
        
        current_track = self.get_current()
        if (current_track is not None):
            await node._send({
                LAVALINK_KEY_NODE_OPERATION: LAVALINK_KEY_NODE_OPERATION_PLAYER_PLAY,
                LAVALINK_KEY_GUILD_ID: str(self.guild_id),
                **current_track.un_pack(),
                LAVALINK_KEY_START_TIME: floor(self.position * 1000.0),
                LAVALINK_KEY_PAUSE: self.is_paused(),
            })
        
        
        filters = self._filters
        if filters:
            data = {
                LAVALINK_KEY_NODE_OPERATION: LAVALINK_KEY_NODE_OPERATION_PLAYER_FILTER,
                LAVALINK_KEY_GUILD_ID: str(self.guild_id),
            }
            
            for filter in filters.values():
                data[filter.json_key] = filter.to_data()
            
            await node._send(data)
        
        
        modified_bands = None
        bands = self.get_bands()
        for band in range(LAVALINK_BAND_COUNT):
            gain = bands[band]
            if gain:
                if modified_bands is None:
                    modified_bands = []
                
                modified_bands.append((band, gain))
        
        
        if (modified_bands is not None):
            await self._set_gains(modified_bands)
    
    
    def __repr__(self):
        """Returns the representation of the player."""
        repr_parts = [
            '<', self.__class__.__name__,
            ' guild_id = ', repr(self.guild_id),
        ]
        
        channel_id = self.channel_id
        if channel_id:
            repr_parts.append(', channel_id = ')
            repr_parts.append(repr(channel_id))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two players are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self is not other:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the player's hash value."""
        return self.guild_id
    
    
    @property
    def channel(self):
        """
        Returns the player's channel.
        
        Returns
        -------
        channel : ``Channel``
        """
        return create_partial_channel_from_id(self.channel_id, ChannelType.unknown, self.guild_id)
    
    
    @property
    def guild(self):
        """
        Returns the player's guild.
        
        Returns
        -------
        guild : ``None | Guild``
        """
        return GUILDS.get(self.guild_id, None)



__all__ = ('EventHandlerPlugin', )

from .meta import EventHandlerPluginMeta


class EventHandlerPlugin(metaclass = EventHandlerPluginMeta):
    """
    Inherit event handler manager plugins from this class, like:
    
    ```py
    class MyPlugin(EventHandlerPlugin):
        my_event = Event(...)
    ```
    
    Each ``Event`` will be picked up and added as a plugin event.
    
    To register the plugin:
    
    ```
    my_plugin = MyPlugin()
    client.events.register_plugin(my_plugin)
    ```
    
    After it, you are able to add new event handlers to your client.
    
    ```py
    @client.events
    async def my_event(...):
        pass
    ```
    
    To ensure an event, do:
    
    ```py
    Task(KOKORO, my_plugin.my_event(...))
    ```
    """
    __slots__ = ()



__all__ = ()

from ...field_putters import float_putter_factory

from ..attachment.fields import (
    put_description, put_waveform, validate_description, validate_duration, validate_name, validate_waveform
)


put_duration = float_putter_factory('duration_secs')



import sys
from os import getcwd as get_current_work_directory
from os.path import (
    dirname as get_directory_name, expanduser as get_user_home_directory, join as join_paths,
    normpath as normalize_path, realpath as get_real_path
)


try:
    from . import __package__ as PACKAGE_NAME
except ImportError:
    # If we have hata not setupped
    PACKAGE_NAME = sys.path[0]
    
    sys.path.append(
        normalize_path(
            join_paths(
                get_directory_name(
                    get_real_path(
                        join_paths(
                            get_current_work_directory(),
                            get_user_home_directory(__file__),
                        )
                    )
                ),
                '..',
            )
        )
    )


PACKAGE = __import__(PACKAGE_NAME)

MAIN = __import__(f'{PACKAGE_NAME}.main').main

__main__ = MAIN.execute_command_from_system_parameters
MAIN.maybe_find_commands()


if __name__ == '__main__':
    __main__()



__all__ = ()

from ..guild.constants import DESCRIPTION_LENGTH_MAX, NAME_LENGTH_MIN, NAME_LENGTH_MAX
from ..guild_badge.constants import (
    BADGE_TAG_LENGTH_MIN as BADGE_TAG_LENGTH_MIN, BADGE_TAG_LENGTH_MAX as BADGE_TAG_LENGTH_MAX
)



__all__ = ()

from ...field_parsers import entity_id_parser_factory, nullable_entity_array_parser_factory
from ...field_putters import entity_id_putter_factory, nullable_entity_array_putter_factory
from ...field_validators import entity_id_validator_factory, nullable_entity_array_validator_factory

from ..guild import Guild


# entitlements

parse_entitlements = nullable_entity_array_parser_factory(
    'entitlements', NotImplemented, include = 'Entitlement'
)
put_entitlements = nullable_entity_array_putter_factory(
    'entitlements', NotImplemented, include = 'Entitlement', force_include_internals = True
)
validate_entitlements = nullable_entity_array_validator_factory(
    'entitlements', NotImplemented, include = 'Entitlement'
)


# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id = entity_id_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', Guild)



__all__ = ('ClientUserBase',)

from re import I as re_ignore_case, compile as re_compile, escape as re_escape

from scarletio import copy_docs, export

from ...color import Color
from ...core import GUILDS, USERS

from ..guild_profile import GuildProfile

from .fields import parse_id, validate_bot
from .flags import UserFlag
from .helpers import _try_get_guild_and_id, _try_get_guild_id
from .orin_user_base import OrinUserBase


@export
class ClientUserBase(OrinUserBase):
    """
    Base class for discord users and clients.
    
    Attributes
    ----------
    avatar_hash : `int`
        The user's avatar's hash in `uint128`.
    
    avatar_type : ``IconType``
        The user's avatar's type.
    
    avatar_decoration : ``None | AvatarDecoration``
        The user's avatar decorations.
    
    banner_color : `None`, ``Color``
        The user's banner color if has any.
    
    banner_hash : `int`
        The user's banner's hash in `uint128`.
    
    banner_type : ``IconType``
        The user's banner's type.
    
    bot : `bool`
        Whether the user is a bot or a user account.
    
    discriminator : `int`
        The user's discriminator. Given to avoid overlapping names.
    
    display_name : `None | str`
        The user's non-unique display name.
    
    flags : ``UserFlag``
        The user's user flags.
    
    guild_profiles : ``dict<int, GuildProfile>``
        A dictionary, which contains the user's guild profiles. If a user is member of a guild, then it should
        have a respective guild profile accordingly.
    
    id : `int`
        The user's unique identifier number.
    
    name : str
        The user's name.
    
    name_plate : ``None | NamePlate``
        The user's name plate.
    
    primary_guild_badge : `None | GuildBadge`
        The user's primary guild's badge.
    
    thread_profiles : ``None | dict<int, ThreadProfile>``
        A Dictionary which contains the thread profiles for the user in thread channel - thread profile relation.
        Defaults to `None`.
    """
    __slots__ = ('bot', 'guild_profiles', 'thread_profiles')
    
    
    def __new__(
        cls,
        *,
        avatar = ...,
        avatar_decoration = ...,
        banner = ...,
        banner_color = ...,
        bot = ...,
        discriminator = ...,
        display_name = ...,
        flags = ...,
        name = ...,
        name_plate = ...,
        primary_guild_badge = ...,
    ):
        """
        Creates a new partial user with the given fields.
        
        Parameters
        ----------
        avatar : ``None | str | bytes-like | Icon``, Optional (Keyword only)
            The user's avatar.
        
        avatar_decoration : ``None | AvatarDecoration``, Optional (Keyword only)
            The user's avatar decoration.
        
        banner : ``None | str | bytes-like | Icon``, Optional (Keyword only)
            The user's banner.
        
        banner_color : `None`, ``Color``, `int`, Optional (Keyword only)
            The user's banner color.
        
        bot : `bool`, Optional (Keyword only)
            Whether the user is a bot or a user account.
        
        discriminator : `str`, `int`, Optional (Keyword only)
            The user's discriminator.
        
        display_name : `None | str`, Optional (Keyword only)
            The user's non-unique display name.
        
        flags : `int`, ``UserFlag``, Optional (Keyword only)
            The user's flags.
        
        name : `str`, Optional (Keyword only)
            The user's name.
        
        name_plate : ``None | NamePlate``, Optional (Keyword only)
            The user's name plate.
        
        primary_guild_badge : `None | GuildBadge`, Optional (Keyword only)
            The user's primary guild's badge.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # bot
        if bot is ...:
            bot = False
        else:
            bot = validate_bot(bot)
        
        self = OrinUserBase.__new__(
            cls,
            avatar = avatar,
            avatar_decoration = avatar_decoration,
            banner = banner,
            banner_color = banner_color,
            discriminator = discriminator,
            display_name = display_name,
            flags = flags,
            name = name,
            name_plate = name_plate,
            primary_guild_badge = primary_guild_badge,
        )
        self.bot = bot
        self.guild_profiles = {}
        self.thread_profiles = None
        return self
    
    
    @classmethod
    def from_data(cls, user_data, guild_profile_data = None, guild_id = 0, *, strong_cache = True):
        """
        Creates a new user from the given data.
        
        Parameters
        ----------
        user_data : `dict<str, object>`
            User data.
        
        guild_profile_data : `None`, `dict<str, object>` = `None`, Optional
            Guild profile data.
        
        guild_id : `int` = `0`, Optional
            The guild's identifier to which the guild profile is bound to.
        
        strong_cache : `bool` = `True`, Optional (Keyword only)
            Whether the instance should be put into its strong cache.
        
        Returns
        -------
        self : `instance<cls>`
        """
        raise NotImplementedError(
            f'`{cls.__name__}` does not support `.from_data` operation, please call it on a sub-type of it.'
        )
    
    
    @classmethod
    @copy_docs(OrinUserBase._from_data_and_difference_update_profile)
    def _from_data_and_difference_update_profile(cls, data, guild):
        user_data = data['user']
        user_id = parse_id(user_data)
        
        self = USERS.get(user_id, None)
        if self is None:
            self = cls.from_data(user_data, data, guild.id)
            old_attributes = None
        
        else:
            old_attributes = self._difference_update_profile(data, guild)
        
        return self, old_attributes
    
    
    @copy_docs(OrinUserBase._difference_update_profile)
    def _difference_update_profile(self, data, guild):
        guild_profile = self.guild_profiles.get(guild.id, None)
        if guild_profile is None:
            self.guild_profiles[guild.id] = GuildProfile.from_data(data)
            guild.users[self.id] = self
            return None
        
        guild_profile._set_joined(data)
        return guild_profile._difference_update_attributes(data)
    
    
    @classmethod
    @copy_docs(OrinUserBase._from_data_and_update_profile)
    def _from_data_and_update_profile(cls, data, guild):
        user_data = data['user']
        user_id = parse_id(user_data)
        
        self = USERS.get(user_id, None)
        if self is None:
            self = cls.from_data(user_data, data, guild.id)
        else:
            self._update_profile(data, guild)
        
        return self
    
    
    @copy_docs(OrinUserBase._update_profile)
    def _update_profile(self, data, guild):
        guild_profile = self.guild_profiles.get(guild.id, None)
        if guild_profile is None:
            self.guild_profiles[guild.id] = GuildProfile.from_data(data)
            guild.users[self.id] = self
            return False
        
        guild_profile._set_joined(data)
        guild_profile._update_attributes(data)
        return True
    
    
    @staticmethod
    def _bypass_no_cache(data, guild_profile_data, guild_id):
        """
        Sets a ``Client``'s guild profile.
        
        > Only available when user or presence caching is disabled.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received user data.
        
        guild_profile_data : `dict<str, object>`
            The user's guild profile's data.
        
        guild_id : `int`
            A respective guild's identifier from where the user data was received.
            Picked up if the given data includes guild member data as well.
        
        Returns
        -------
        user : ``ClientUserBase``
        """
        user_id = parse_id(data)
        
        try:
            user = USERS[user_id]
        except KeyError:
            return
        
        try:
            guild_profile = user.guild_profiles[guild_id]
        except KeyError:
            user.guild_profiles[guild_id] = GuildProfile.from_data(guild_profile_data)
        else:
            guild_profile._set_joined(guild_profile_data)
            guild_profile._update_attributes(guild_profile_data)
        
        return user
    
    
    @classmethod
    def _from_client(cls, client, include_internals):
        """
        Creates a client alter ego.
        
        Parameters
        ----------
        client : ``Client``
            The client to copy.
        include_internals : `bool`
            Whether internal fields should be copied as well.
        
        Returns
        -------
        user : ``ClientUserBase``
        """
        self = object.__new__(cls)
        self.avatar_hash = client.avatar_hash
        self.avatar_type = client.avatar_type
        
        # avatar_decoration
        avatar_decoration = client.avatar_decoration
        if (avatar_decoration is not None):
            avatar_decoration = avatar_decoration.copy()
        self.avatar_decoration = avatar_decoration
        
        self.banner_color = client.banner_color
        self.banner_hash = client.banner_hash
        self.banner_type = client.banner_type
        self.bot = client.bot
        self.discriminator = client.discriminator
        self.display_name = client.display_name
        self.flags = client.flags
        
        if include_internals:
            guild_profiles = client.guild_profiles
            if (guild_profiles is not None):
                guild_profiles = {guild_id: guild_profile.copy() for guild_id, guild_profile in guild_profiles.items()}
        else:
            guild_profiles = {}
        
        self.guild_profiles = guild_profiles
        
        if include_internals:
            user_id = client.id
        else:
            user_id = 0
        self.id = user_id
        self.name = client.name
        
        # name_plate
        name_plate = client.name_plate
        if (name_plate is not None):
            name_plate = name_plate.copy()
        self.name_plate = name_plate
        
        primary_guild_badge = client.primary_guild_badge
        if (primary_guild_badge is not None):
            primary_guild_badge = primary_guild_badge.copy()
        self.primary_guild_badge = primary_guild_badge
        
        if include_internals:
            thread_profiles = client.thread_profiles
            if (thread_profiles is not None):
                thread_profiles = {
                    channel_id: thread_profile.copy() for channel_id, thread_profile in thread_profiles.items()
                }
        else:
            thread_profiles = None
        self.thread_profiles = thread_profiles
        
        return self
    
    
    @copy_docs(OrinUserBase._set_default_attributes)
    def _set_default_attributes(self):
        OrinUserBase._set_default_attributes(self)
        
        self.bot = False
        self.guild_profiles = {}
        self.thread_profiles = None
    
    
    @copy_docs(OrinUserBase.copy)
    def copy(self):
        new = OrinUserBase.copy(self)
        new.bot = self.bot
        new.guild_profiles = {}
        new.thread_profiles = None
        return new
    
    
    def copy_with(
        self,
        *,
        avatar = ...,
        avatar_decoration = ...,
        banner = ...,
        banner_color = ...,
        bot = ...,
        discriminator = ...,
        display_name = ...,
        flags = ...,
        name = ...,
        name_plate = ...,
        primary_guild_badge = ...,
    ):
        """
        Copies the user with the given fields.
        
        Parameters
        ----------
        avatar : ``None | str | bytes-like | Icon``, Optional (Keyword only)
            The user's avatar.
        
        avatar_decoration : ``None | AvatarDecoration``, Optional (Keyword only)
            The user's avatar decoration.
        
        banner : ``None | str | bytes-like | Icon``, Optional (Keyword only)
            The user's banner.
        
        banner_color : `None`, ``Color``, `int`, Optional (Keyword only)
            The user's banner color.
        
        bot : `bool`, Optional (Keyword only)
            Whether the user is a bot or a user account.
        
        discriminator : `str`, `int`, Optional (Keyword only)
            The user's discriminator.
        
        display_name : `None | str`, Optional (Keyword only)
            The user's non-unique display name.
        
        flags : `int`, ``UserFlag``, Optional (Keyword only)
            The user's flags.
        
        name : `str`, Optional (Keyword only)
            The user's name.
        
        name_plate : ``None | NamePlate``, Optional (Keyword only)
            The user's name plate.
        
        primary_guild_badge : `None | GuildBadge`, Optional (Keyword only)
            The user's primary guild's badge.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # bot
        if bot is ...:
            bot = self.bot
        else:
            bot = validate_bot(bot)
        
        # Construct
        new = OrinUserBase.copy_with(
            self,
            avatar = avatar,
            avatar_decoration = avatar_decoration,
            banner = banner,
            banner_color = banner_color,
            discriminator = discriminator,
            display_name = display_name,
            flags = flags,
            name = name,
            name_plate = name_plate,
            primary_guild_badge = primary_guild_badge,
        )
        new.bot = bot
        new.guild_profiles = {}
        new.thread_profiles = None
        return new
    
    
    @copy_docs(OrinUserBase._get_hash_partial)
    def _get_hash_partial(self):
        hash_value = OrinUserBase._get_hash_partial(self)
        
        # bot
        hash_value ^= self.bot << 34
        
        # guild_profiles | internal -> ignore
        # thread_profiles | internal -> ignore
        
        return hash_value
    
    
    @copy_docs(OrinUserBase._delete)
    def _delete(self):
        # we cannot full delete a user, because of the mentions, so we delete it only from the guilds
        guild_profiles = self.guild_profiles
        while guild_profiles:
            guild_id, guild_profile = guild_profiles.popitem()
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                continue
            
            try:
                del guild.users[self.id]
            except KeyError:
                pass
        
        # TODO: Should we delete the user from threads too?
    
    
    @copy_docs(OrinUserBase.color_at)
    def color_at(self, guild):
        guild_id = _try_get_guild_id(guild)
        
        try:
            guild_profile = self.guild_profiles[guild_id]
        except KeyError:
            pass
        else:
            return guild_profile.color
        
        return Color()
    
    
    @copy_docs(OrinUserBase.name_at)
    def name_at(self, guild):
        guild_id = _try_get_guild_id(guild)
        
        try:
            guild_profile = self.guild_profiles[guild_id]
        except KeyError:
            pass
        else:
            nick = guild_profile.nick
            if (nick is not None):
                return nick
        
        display_name = self.display_name
        if (display_name is not None):
            return display_name
        
        return self.name
    
    
    @copy_docs(OrinUserBase.has_name_like_at)
    def has_name_like_at(self, name, guild):
        if name.startswith('@'):
            name = name[1:]
        
        name_length = len(name)
        if (name_length < 1):
            return False
        
        if name_length > 5:
            if name_length > 37:
                return False
            
            if name[-5] == '#':
                try:
                    discriminator = int(name[-4:])
                except ValueError:
                    pass
                else:
                    stripped_name = name[:-5]
                    if (self.discriminator == discriminator) and (self.name == stripped_name):
                        return True
        
        if name_length > 32:
            return False
        
        pattern = re_compile(re_escape(name), re_ignore_case)
        
        if pattern.search(self.name) is not None:
            return True
        
        display_name = self.display_name
        if (display_name is not None) and (pattern.search(display_name) is not None):
            return True
        
        guild_id = _try_get_guild_id(guild)
        try:
            guild_profile = self.guild_profiles[guild_id]
        except KeyError:
            pass
        else:
            nick = guild_profile.nick
            if (nick is not None):
                if pattern.search(nick) is not None:
                    return True
        
        return False
    
    
    @copy_docs(OrinUserBase.has_role)
    def has_role(self, role):
        guild_id = role.guild_id
        
        try:
            guild_profile = self.guild_profiles[guild_id]
        except KeyError:
            return False
        
        role_id = role.id
        if guild_id == role_id:
            return True
        
        role_ids = guild_profile.role_ids
        if role_ids is None:
            return False
        
        if role_id in role_ids:
            return True
        
        return False
    
    
    @copy_docs(OrinUserBase.top_role_at)
    def top_role_at(self, guild, default = None):
        guild, guild_id = _try_get_guild_and_id(guild)
        
        try:
            guild_profile = self.guild_profiles[guild_id]
        except KeyError:
            pass
        else:
            top_role = guild_profile.get_top_role()
            if (top_role is not None):
                return top_role
            
            if (guild is not None):
                top_role = guild.default_role
                if (top_role is not None):
                    return top_role
        
        return default
    
    
    @copy_docs(OrinUserBase.can_use_emoji)
    def can_use_emoji(self, emoji):
        if emoji.is_unicode_emoji():
            return True
        
        if not emoji.available:
            return False
        
        guild_id = emoji.guild_id
        
        try:
            guild_profile = self.guild_profiles[guild_id]
        except KeyError:
            return False
        
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            return False
        
        # Already deleted?
        if emoji.id not in guild.emojis:
            return False
        
        emoji_role_ids = emoji.role_ids
        if (emoji_role_ids is not None):
            guild_profile_role_ids = guild_profile.role_ids
            if (guild_profile_role_ids is None):
                return False
            
            if not {*emoji_role_ids} & {*guild_profile_role_ids}:
                return False
        
        return True
    
    
    @copy_docs(OrinUserBase.has_higher_role_than)
    def has_higher_role_than(self, role):
        guild_id = role.guild_id
        
        try:
            guild_profile = self.guild_profiles[guild_id]
        except KeyError:
            return False
        
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            pass
        else:
            if guild.owner_id == self.id:
                return True
        
        top_role = guild_profile.get_top_role()
        if top_role is None:
            return False
        
        if top_role > role:
            return True
        
        return False
    
    
    @copy_docs(OrinUserBase.has_higher_role_than_at)
    def has_higher_role_than_at(self, user, guild):
        if self is user:
            return False
        
        guild, guild_id = _try_get_guild_and_id(guild)
        if (not guild_id):
            return False
        
        try:
            own_profile = self.guild_profiles[guild_id]
        except KeyError:
            return False
        
        if (guild is not None) and (guild.owner_id == self.id):
            return True
        
        try:
            other_profile = user.guild_profiles[guild_id]
        except KeyError:
            # We always have higher permissions if the other user is not in the guild or if it is a webhook.
            return True
        
        if (guild is not None) and (guild.owner_id == user.id):
            return False
        
        own_top_role = own_profile.get_top_role()
        if own_top_role is None:
            return False
        
        other_top_role = other_profile.get_top_role()
        if other_top_role is None:
            return True
        
        if own_top_role > other_top_role:
            return True
        
        return False
    
    
    @copy_docs(OrinUserBase.get_guild_profile_for)
    def get_guild_profile_for(self, guild):
        guild_id = _try_get_guild_id(guild)
        return self.guild_profiles.get(guild_id, None)
    
    
    @copy_docs(OrinUserBase.iter_guilds_and_profiles)
    def iter_guilds_and_profiles(self):
        for guild_id, guild_profile in self.guild_profiles.items():
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                continue
            
            yield guild, guild_profile
    
    
    @copy_docs(OrinUserBase.iter_guilds)
    def iter_guilds(self):
        for guild_id in self.guild_profiles.keys():
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                continue
            
            yield guild
    
    
    @copy_docs(OrinUserBase.is_boosting)
    def is_boosting(self, guild):
        guild_id = _try_get_guild_id(guild)
        
        try:
            guild_profile = self.guild_profiles[guild_id]
        except KeyError:
            return False
        
        return (guild_profile.boosts_since is not None)
    
    
    @property
    @copy_docs(OrinUserBase.partial)
    def partial(self):
        for guild_id in self.guild_profiles.keys():
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                continue
            
            if guild.partial:
                continue
            
            return False
        
        return True



__all__ = ()

from ...stage.stage.fields import validate_channel_id

from ..audit_log_entry_detail_conversion import AuditLogEntryDetailConversion, AuditLogEntryDetailConversionGroup
from ..conversion_helpers.converters import value_deserializer_id, value_serializer_id


# ---- channel_id ----

CHANNEL_ID_CONVERSION = AuditLogEntryDetailConversion(
    'channel_id',
    'channel_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_channel_id,
)


# ---- Construct ----

STAGE_CONVERSIONS = AuditLogEntryDetailConversionGroup(
    CHANNEL_ID_CONVERSION,
)



__all__ = ()

from ...field_parsers import nullable_string_parser_factory
from ...field_putters import nullable_string_optional_putter_factory
from ...field_validators import nullable_string_validator_factory

# join

parse_join = nullable_string_parser_factory('join')
put_join = nullable_string_optional_putter_factory('join')
validate_join = nullable_string_validator_factory('join', 0, 1024)

# match

parse_match = nullable_string_parser_factory('match')
put_match = nullable_string_optional_putter_factory('match')
validate_match = nullable_string_validator_factory('match', 0, 1024)

# spectate

parse_spectate = nullable_string_parser_factory('spectate')
put_spectate = nullable_string_optional_putter_factory('spectate')
validate_spectate = nullable_string_validator_factory('spectate', 0, 1024)



__all__ = ()

from os.path import join as join_paths

from ....core import LIBRARY_CALLED_DIRECTLY, register

from .helpers import _validate_bot, _validate_layout, _validate_project_name, _validate_name
from .layouts import DEFAULT_LAYOUT, LAYOUT_DESCRIPTIONS, get_project_structure_builder
from .naming import get_project_module_name


def _build_scaffold_description():
    """
    Builds description of the `scaffold` command.
    
    Returns
    -------
    description : `str`
    """
    description_parts = []
    
    description_parts.append(
        'Scaffold can be used to automatically create a basic structure for your application.\n'
        'This includes code, configuration files, and other necessary files and assets.\n'
        '\n'
        'Available layouts:'
    )
    for index, (layout_name, layout_description) in enumerate(sorted(LAYOUT_DESCRIPTIONS.items()), 1):
        description_parts.append('\n\n')
        index_representation = str(index)
        description_parts.append(index_representation)
        description_parts.append('.: ')
        description_parts.append(layout_name.upper())
        
        underscore_length = len(index_representation) + 4 + len(layout_name)
        if layout_name == DEFAULT_LAYOUT:
            description_parts.append(' (default)')
            underscore_length += 10
        
        description_parts.append(':\n')
        description_parts.append('-' * underscore_length)
        description_parts.append('\n')
        
        description_parts.append(layout_description)
    
    description_parts.append('\n')
    return ''.join(description_parts)


@register(
    available = LIBRARY_CALLED_DIRECTLY,
    description = _build_scaffold_description(),
)
def scaffold(
    name : str,
    *bot : str,
    project_name : str = None,
    layout : str = None,
):
    """
    Scaffolds a project.
    """
    directory_path, error_message = _validate_name(name)
    if error_message is not None:
        return error_message
    
    bot_names, error_message = _validate_bot(bot)
    if error_message is not None:
        return error_message
    
    project_name, error_message = _validate_project_name(project_name, name)
    if error_message is not None:
        return error_message
    
    layout, error_message = _validate_layout(layout)
    if error_message is not None:
        return error_message
    
    project_module_name = get_project_module_name(project_name)
    create_project_structure = get_project_structure_builder(layout)
    create_project_structure(directory_path, project_module_name, bot_names)
    return (
        f'Project {project_module_name} created.\n'
        f'Please check {join_paths(directory_path, "README.md")} about whats next.\n'
    )



__all__ = ()

from ...field_parsers import (
    entity_id_parser_factory, force_string_parser_factory, nullable_entity_array_parser_factory
)
from ...field_putters import (
    entity_id_putter_factory, force_string_putter_factory, nullable_entity_array_putter_factory
)
from ...field_validators import (
    entity_id_validator_factory, force_string_validator_factory, nullable_entity_array_validator_factory
)
from ...user import ClientUserBase

from ..team_member import TeamMember

from .constants import NAME_LENGTH_MAX, NAME_LENGTH_MIN

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('team_id')

# members

parse_members = nullable_entity_array_parser_factory('members', TeamMember)
put_members = nullable_entity_array_putter_factory('members', TeamMember)
validate_members = nullable_entity_array_validator_factory('members', TeamMember)

# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)

# owner_id

parse_owner_id = entity_id_parser_factory('owner_user_id')
put_owner_id = entity_id_putter_factory('owner_user_id')
validate_owner_id = entity_id_validator_factory('owner_id', ClientUserBase)



__all__ = ('User',)

from warnings import warn

from scarletio import copy_docs

from ....env import CACHE_PRESENCE, CACHE_USER

from ...core import GUILDS, USERS
from ...precreate_helpers import process_precreate_parameters_and_raise_extra

from ..guild_profile import GuildProfile
from ..status_by_platform import Status, StatusByPlatform

from .client_user_base import ClientUserBase
from .client_user_presence_base import ClientUserPBase
from .fields import (
    parse_bot, parse_id, validate_activities, validate_avatar_decoration, validate_banner_color, validate_bot,
    validate_discriminator, validate_display_name, validate_flags, validate_id, validate_name, validate_name_plate,
    validate_primary_guild_badge, validate_status, validate_status_by_platform
)
from .flags import UserFlag
from .orin_user_base import USER_BANNER
from .user_base import USER_AVATAR


def _validate_statuses_deprecated(input_value):
    warn(
        (
            f'`statuses` is deprecated and will be removed in 2026 february. '
            f'Please use `status_by_platform` instead.'
        ),
        FutureWarning,
        stacklevel = 3,
    )
    
    if (input_value is None):
        return None
    
    if isinstance(input_value, dict):
        if not input_value:
            return None
        
        unexpected_keys = {*input_value.keys()} - {'desktop', 'embedded', 'mobile', 'web'}
        if unexpected_keys:
            raise ValueError(
                f'`statuses contains unexpected keys: {", ".join([repr(key) for key in unexpected_keys])}'
            )
        
        return StatusByPlatform.from_data(input_value)
    
    raise TypeError(
        f'`statuses` can be `None | dict<str, str>`, got {type(input_value).__name__}; {input_value!r}.'
    )


PRECREATE_FIELDS = {
    'avatar': ('avatar', USER_AVATAR.validate_icon),
    'avatar_decoration': ('avatar_decoration', validate_avatar_decoration),
    'banner': ('banner', USER_BANNER.validate_icon),
    'banner_color': ('banner_color', validate_banner_color),
    'bot': ('bot', validate_bot),
    'discriminator': ('discriminator', validate_discriminator),
    'display_name': ('display_name', validate_display_name),
    'flags': ('flags', validate_flags),
    'name': ('name', validate_name),
    'name_plate': ('name_plate', validate_name_plate),
    'primary_guild_badge': ('primary_guild_badge', validate_primary_guild_badge),
}

if not CACHE_PRESENCE:
    USER_BASE_TYPE = ClientUserBase

else:
    USER_BASE_TYPE = ClientUserPBase
    
    PRECREATE_FIELDS = {
        **PRECREATE_FIELDS,
        'activities': ('activities', validate_activities),
        'status': ('status', validate_status),
        'status_by_platform': ('status_by_platform', validate_status_by_platform),
        'statuses': ('status_by_platform', _validate_statuses_deprecated),
    }


class User(USER_BASE_TYPE):
    """
    Represents a Discord user.
    
    Attributes
    ----------
    activities : ``None | list<Activity>``
        A list of the client's activities. Defaults to `None`
        
        > Only available if presence caching is enabled.
    
    avatar_hash : `int`
        The user's avatar's hash in `uint128`.
    
    avatar_type : ``IconType``
        The user's avatar's type.
    
    avatar_decoration : ``None | AvatarDecoration``
        The user's avatar decorations.
    
    banner_color : `None`, ``Color``
        The user's banner color if has any.
    
    banner_hash : `int`
        The user's banner's hash in `uint128`.
    
    banner_type : ``IconType``
        The user's banner's type.
    
    bot : `bool`
        Whether the user is a bot or a user account.
    
    discriminator : `int`
        The user's discriminator. Given to avoid overlapping names.
    
    display_name : `None`, `str`
        The user's non-unique display name.
    
    flags : ``UserFlag``
        The user's user flags.
    
    guild_profiles : ``dict<int, GuildProfile>``
        A dictionary, which contains the user's guild profiles. If a user is member of a guild, then it should
        have a respective guild profile accordingly.
    
    id : `int`
        The user's unique identifier number.
    
    name : `str`
        The user's name.
    
    name_plate : ``None | NamePlate``
        The user's name plate.
    
    primary_guild_badge : ``None | GuildBadge``
        The user's primary guild's badge.
    
    status : ``Status``
        The user's display status.
        
        > Only available if presence caching is enabled.
    
    status_by_platform : ``None | StatusByPlatform`
        The user's status for each platform.
        
        > Only available if presence caching is enabled.
    
    thread_profiles : ``None | dict<int, ThreadProfile>``
        A Dictionary which contains the thread profiles for the user in thread channel - thread profile relation.
        Defaults to `None`.
    
    Notes
    -----
    Instances of this class are weakreferable.
    """
    __slots__ = ()
    
    if CACHE_PRESENCE:
        @classmethod
        @copy_docs(USER_BASE_TYPE.from_data)
        def from_data(cls, user_data, guild_profile_data = None, guild_id = 0, *, strong_cache = True):
            user_id = parse_id(user_data)
            
            try:
                self = USERS[user_id]
            except KeyError:
                self = object.__new__(cls)
                self.id = user_id
                self.guild_profiles = {}
                self.thread_profiles = None
                self.status = Status.offline
                self.status_by_platform = None
                self.activities = None
                
                USERS[user_id] = self
            
            self.bot = parse_bot(user_data)
            self._update_attributes(user_data)
            
            if (guild_profile_data is not None) and guild_id:
                try:
                    guild_profile = self.guild_profiles[guild_id]
                except KeyError:
                    self.guild_profiles[guild_id] = GuildProfile.from_data(guild_profile_data)
                else:
                    guild_profile._set_joined(guild_profile_data)
                    guild_profile._update_attributes(guild_profile_data)
                
                if strong_cache:
                    try:
                        guild = GUILDS[guild_id]
                    except KeyError:
                        pass
                    else:
                        guild.users[user_id] = self
            
            return self
    
    elif CACHE_USER:
        @classmethod
        @copy_docs(USER_BASE_TYPE.from_data)
        def from_data(cls, user_data, guild_profile_data = None, guild_id = 0, *, strong_cache = True):
            user_id = parse_id(user_data)

            try:
                self = USERS[user_id]
            except KeyError:
                self = object.__new__(cls)
                self.id = user_id
                self.guild_profiles = {}
                self.thread_profiles = None
                
                USERS[user_id] = self
            
            self.bot = parse_bot(user_data)
            self._update_attributes(user_data)
            
            if (guild_profile_data is not None) and guild_id:
                try:
                    guild_profile = self.guild_profiles[guild_id]
                except KeyError:
                    self.guild_profiles[guild_id] = GuildProfile.from_data(guild_profile_data)
                else:
                    guild_profile._set_joined(guild_profile_data)
                    guild_profile._update_attributes(guild_profile_data)
                
                if strong_cache:
                    try:
                        guild = GUILDS[guild_id]
                    except KeyError:
                        pass
                    else:
                        guild.users[user_id] = self
            
            return self
    
    else:
        @classmethod
        @copy_docs(USER_BASE_TYPE.from_data)
        def from_data(cls, user_data, guild_profile_data = None, guild_id = 0, strong_cache = True):
            user_id = parse_id(user_data)
            
            try:
                self = USERS[user_id]
            except KeyError:
                self = object.__new__(cls)
                self.id = user_id
                self.guild_profiles = {}
                self.thread_profiles = None
                
                USERS[user_id] = self
            
            self.bot = parse_bot(user_data)
            self._update_attributes(user_data)
            
            if (guild_profile_data is not None) and guild_id:
                self.guild_profiles[guild_id] = GuildProfile.from_data(guild_profile_data)
            
            # Do we want to strong cache at this case?
            # if strong_cache:
            #     try:
            #         guild = GUILDS[guild_id]
            #     except KeyError:
            #         pass
            #     else:
            #         guild.users[user_id] = self
            
            return self
    
    
    @classmethod
    def precreate(cls, user_id, **keyword_parameters):
        """
        Precreates a user by creating a partial one with the given parameters. When the user is loaded, the precreated
        one is picked up and is updated. If an already existing user would be precreated, returns that instead of
        creating a new one, and updates it only, if it is still a partial one.
        
        Parameters
        ----------
        user_id : `int`, `str`
            The user's id.
        
        **keyword_parameters : keyword parameters
            Additional predefined attributes for the user.
        
        Other Parameters
        ----------------
        activities : `iterable`, `list` of ``Activity``, Optional (Keyword only)
            A list of the client's activities.
            
            > Only available if presence caching is enabled.
        
        avatar : ``None | str | Icon``, Optional (Keyword only)
            The user's avatar.
        
        avatar_decoration : ``None | AvatarDecoration``, Optional (Keyword only)
            The user's avatar decoration.
        
        name_plate : ``None | NamePlate``, Optional (Keyword only)
            The user's name plate.
        
        banner : ``None | str | Icon``, Optional (Keyword only)
            The user's banner.
            
            > Mutually exclusive with `banner_type` and `banner_hash`.
        
        banner_color : `None`, ``Color``
            The user's banner color.
        
        bot : `bool`, Optional (Keyword only)
            Whether the user is a bot account.
        
        flags : ``UserFlag``, `int`, Optional (Keyword only)
            The user's ``.flags``. If not passed as ``UserFlag``, then will be converted to it.
        
        name : `str`, Optional (Keyword only)
            The user's ``.name``.
        
        primary_guild_badge : ``None | GuildBadge``, Optional (Keyword only)
            The user's primary guild's badge.
        
        discriminator : `int`, `str`, Optional (Keyword only)
            The user's discriminator.
        
        display_name : `None`, `str`, Optional (Keyword only)
            The user's non-unique display name.
        
        status : `Status`, `str`, Optional (Keyword only)
            The user's display status.
        
            > Only available if presence caching is enabled.
        
        status_by_platform : ``None | StatusByPlatform`, Optional (Keyword only)
            The user's status for each platform.
        
            > Only available if presence caching is enabled.
        
        Returns
        -------
        user : ``ClientUserBase``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        user_id = validate_id(user_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = USERS[user_id]
        except KeyError:
            self = cls._create_empty(user_id)
            USERS[user_id] = self
        
        else:
            if not self.partial:
                return self
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self



from .attachment_input import *
from .base import *
from .button import *
from .channel_select import *
from .container import *
from .fields import *
from .label import *
from .mentionable_select import *
from .role_select import *
from .row import *
from .section import *
from .string_select import *
from .text_input import *
from .user_select import *


__all__ = (
    *attachment_input.__all__,
    *base.__all__,
    *button.__all__,
    *channel_select.__all__,
    *container.__all__,
    *fields.__all__,
    *label.__all__,
    *mentionable_select.__all__,
    *role_select.__all__,
    *row.__all__,
    *section.__all__,
    *string_select.__all__,
    *text_input.__all__,
    *user_select.__all__,
)
